"""Helper functions that work with Voronoi data structures.

This module provides functions to help work with the Voronoi data
structures and to map them to more sensible data structures.

The terminology used within this module can sometimes be confusing.

region :
    Polygon definied by Voronoi. These polygons can be unbound (infinite in
    area).
vertex :
    The points that define the Voronoi regions.
ridge :
    Lines that join adjacent vertices. Ridges are the edges of Voronoi
    regions.
point :
    Each region has an associated point, which lies inside the region. Each
    region has only one point, and each point has only on region.
patch :
    Similar to a region but is always bound.
link :
    Similar to a ridge but is always part of at least one patch.
node :
    Similar to a vertex but bounds only links.
corner :
    Same as a point.
"""
from six.moves import range
import numpy as np

from ..sort.sort import remap
from ...core.utils import as_id_array


def flatten_vertices_at_region(regions):
    """
    Examples
    --------
    >>> from scipy.spatial import Voronoi
    >>> from landlab.graph.voronoi.voronoi_helpers import flatten_vertices_at_region
    >>> points = [[0. , 0.], [1. , 0.], [2. , 0.],
    ...           [0.1, 1.], [1.1, 1.], [2.1, 1.],
    ...           [0.2, 2.], [1.2, 2.], [2.2, 2.]]
    >>> voronoi = Voronoi(points)
    >>> (vertices, count) = flatten_vertices_at_region(voronoi.regions)
    >>> vertices
    array([ 1,  0, -1,  4,  2, -1,  3,  3, -1,  4,  0, -1,  3,  7,  5, -1,  6,
            5,  2, -1,  6, -1,  7,  1,  0,  4,  2,  5,  7,  1, -1,  6])
    >>> count
    array([3, 0, 4, 2, 4, 4, 3, 2, 6, 4])
    """
    vertices_at_region = np.array(np.concatenate(regions), dtype=int)
    vertices_per_region = np.array(
        [len(region) for region in regions], dtype=int)

    return vertices_at_region, vertices_per_region


class VoronoiConverter(object):
    def __init__(self, voronoi, min_patch_size=3):
        self._voronoi = voronoi
        self._min_patch_size = min_patch_size

    @property
    def n_regions(self):
        return len(self._voronoi.regions)

    @property
    def regions(self):
        return self._voronoi.regions

    @property
    def max_patch_size(self):
        try:
            return self._max_patch_size
        except AttributeError:
            self._max_patch_size = max([len(r) for r in self.regions])
            return self._max_patch_size

    def get_finite_regions(self):
        """Get regions of finite area.

        Examples
        --------
        >>> from scipy.spatial import Voronoi
        >>> from landlab.graph.voronoi.voronoi_helpers import VoronoiConverter

        >>> points = [[0. , 0.], [1. , 0.], [2. , 0.],
        ...           [0.1, 1.], [1.1, 1.], [2.1, 1.],
        ...           [0.2, 2.], [1.2, 2.], [2.2, 2.]]
        >>> voronoi = Voronoi(points)

        >>> converter = VoronoiConverter(voronoi)
        >>> converter.get_finite_regions()
        array([0, 0, 0, 0, 0, 0, 0, 0, 1, 0])
        """
        try:
            return self._is_finite_region
        except AttributeError:
            from .ext.voronoi import _is_finite_region

            n_regions = self.n_regions
            vertices_at_region, vertices_per_region = flatten_vertices_at_region(
                self.regions)

            is_finite_region = np.empty(n_regions, dtype=int)
            _is_finite_region(vertices_at_region, vertices_per_region,
                              is_finite_region, self._min_patch_size)

            self._is_finite_region = is_finite_region
            return self._is_finite_region

    def is_patch(self, region):
        """Check if a voronoi region in bounded.

        Parameters
        ----------
        region : array_like of int
            Identifiers for the vertices of a region.

        Examples
        --------
        >>> from landlab.graph.voronoi.voronoi_helpers import VoronoiConverter

        >>> converter = VoronoiConverter(None)
        >>> converter.is_patch([1, 2, 3])
        True
        >>> converter.is_patch([1, 2, 3, -1])
        False
        >>> converter.is_patch([])
        False

        Specify the minimum number sides for a valid patch.

        >>> converter = VoronoiConverter(None, min_patch_size=4)
        >>> converter.is_patch([1, 2, 3])
        False
        """
        return len(region) >= self._min_patch_size and -1 not in region

    def is_link(self, ridge):
        """Check if a voronoi ridge is a valid link.

        A ridge is a valid link if at least one of its neighbor regions is
        bound.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.
        ridge : int
            Identifier of a ridge in the voronoi graph.

        Returns
        -------
        bool
            True if the ridge is a link.
        """
        for point in self._voronoi.ridge_points[ridge]:
            region = self._voronoi.point_region[point]
            if self.is_patch(self._voronoi.regions[region]):
                return True
        return False

    def get_patch_at_region(self):
        """Map voronoi regions to patches.

        Regions that do not have corresponding patches, are -1.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.

        Returns
        -------
        ndarray of int, shape `(n_regions, 1)`
            Array of patch identifiers for every voronoi region.
        """
        patch_at_region = np.empty(len(self._voronoi.regions), dtype=int)

        patch = 0
        for n, region in enumerate(self._voronoi.regions):
            if self.is_patch(region):
                patch_at_region[n] = patch
                patch += 1
            else:
                patch_at_region[n] = -1

        return patch_at_region

    def get_link_at_ridge(self):
        """Map voronoi ridges to links.

        Ridges that do not have corresponding links, are -1.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.

        Returns
        -------
        ndarray of int, shape `(n_ridges, 1)`
            Array of link identifiers for every voronoi ridge.
        """
        link_at_ridge = np.empty(len(self._voronoi.ridge_vertices), dtype=int)

        link = 0
        for n, ridge in enumerate(self._voronoi.ridge_vertices):
            if -1 not in ridge and self.is_link(n):
                link_at_ridge[n] = link
                link += 1
            else:
                link_at_ridge[n] = -1

        return link_at_ridge

    def get_patches_at_link(self):
        """Get patches on either side of each link.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.

        Returns
        -------
        ndarray of int, shape `(n_links, 2)`
            Array of patch identifiers for every link.
        """
        points_at_ridge = self._voronoi.ridge_points
        region_at_point = self._voronoi.point_region

        regions_at_ridge = as_id_array(region_at_point[points_at_ridge])

        patch_at_region = self.get_patch_at_region()
        link_at_ridge = self.get_link_at_ridge()

        patch_at_ridge = remap(regions_at_ridge, patch_at_region, inplace=True)

        return patch_at_ridge[link_at_ridge >= 0]

    def get_node_at_vertex(self):
        """Map voronoi vertices to nodes.

        Vertices that do not have corresponding nodes, are -1.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.

        Returns
        -------
        ndarray of int, shape `(n_vertices, 1)`
            Array of node identifiers for every voronoi vertex.
        """
        node_at_vertex = np.full(len(self._voronoi.vertices), -1, dtype=int)

        node = 0
        for region in self._voronoi.regions:
            if self.is_patch(region):
                for vertex in region:
                    if node_at_vertex[vertex] == -1:
                        node_at_vertex[vertex] = node
                        node += 1

        return node_at_vertex

    def get_nodes_at_link(self):
        """Get end nodes for links of a voronoi graph.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.

        Returns
        -------
        ndarray of int, shape `(n_links, 2)`
            Array of node identifiers for every link.
        """
        node_at_vertex = self.get_node_at_vertex()
        link_at_ridge = self.get_link_at_ridge()
        vertices_at_ridge = np.array(self._voronoi.ridge_vertices)

        nodes_at_link = node_at_vertex[vertices_at_ridge[link_at_ridge >= 0].flat]
        nodes_at_link.shape = (-1, 2)

        return nodes_at_link

    def get_nodes(self):
        """Get nodes of a voronoi graph.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.

        Returns
        -------
        ndarray of float, shape `(n_nodes, 2)`
            Array of coordinates `(x, y)` for every node.
        """
        node_at_vertex = self.get_node_at_vertex()
        nodes = np.argsort(node_at_vertex[node_at_vertex >= 0])

        return self._voronoi.vertices[node_at_vertex >= 0][nodes]

    def get_ridges_at_region(self):
        """Get ridges that bound each voronoi region.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.

        Returns
        -------
        ndarray of int, shape `(n_regions, max_ridges_per_region)`
            Ridge identifiers that define each region (padded with -1).
        """
        n_regions = self.n_regions
        max_links_per_patch = self.max_patch_size

        ridges_at_region = np.full((n_regions, max_links_per_patch), -1,
                                   dtype=int)

        ridge_at_vertices = {}
        for ridge, vertices in enumerate(self._voronoi.ridge_vertices):
            ridge_at_vertices[tuple(vertices)] = ridge

        for region, vertices in enumerate(self.regions):
            if self.is_patch(vertices):
                for n in range(len(vertices) - 1):
                    pair = [vertices[n], vertices[n + 1]]
                    pair.sort()

                    ridge = ridge_at_vertices[tuple(pair)]

                    ridges_at_region[region, n] = ridge
                pair = [vertices[len(vertices) - 1], vertices[0]]
                pair.sort()

                ridge = ridge_at_vertices[tuple(pair)]

                ridges_at_region[region, len(vertices) - 1] = ridge

        return ridges_at_region

    def get_links_at_patch(self):
        """Get links that bound each patch from a voronoi graph.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.

        Returns
        -------
        ndarray of int, shape `(n_patches, max_links_per_region)`
            Link identifiers that define each patch (padded with -1).
        """
        from ..sort.ext.remap_element import remap_graph_element_ignore

        ridges_at_region = self.get_ridges_at_region()
        link_at_ridge = self.get_link_at_ridge()

        patches = self.get_patch_at_region() >= 0
        links_at_patch = ridges_at_region[patches]

        remap_graph_element_ignore(links_at_patch.reshape((-1, )),
                                   link_at_ridge, -1)
        return links_at_patch

    def get_corner_at_patch(self):
        """Get corner that is contained within each patch from a voronoi graph.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.

        Returns
        -------
        ndarray of int, shape `(n_patches, )`
            Corner identifiers contained in patches.
        """
        finite_regions = self.get_finite_regions()
        n_patches = finite_regions.sum()

        region_at_patch = np.argsort(self.get_patch_at_region())[- n_patches:]
        point_at_region = np.argsort(self._voronoi.point_region)

        return point_at_region[region_at_patch - 1]

    def get_corners_at_link(self):
        """Get corners on either side of links.

        Parameters
        ----------
        voronoi : Voronoi
            A voronoi graph.

        Returns
        -------
        ndarray of int, shape `(n_links, 2)`
            Corner identifiers on either side of each link.
        """
        points_at_ridge = self._voronoi.ridge_points
        link_at_ridge = self.get_link_at_ridge()
        return points_at_ridge[link_at_ridge >= 0]



#! /usr/bin/env python
"""Container that holds a collection of named data-fields."""

import numpy as np


_UNKNOWN_UNITS = '?'


class Error(Exception):

    """Base class for errors in this module."""

    pass


class FieldError(Error, KeyError):

    """Raise this error for a missing field name."""

    def __init__(self, field):
        self._field = field

    def __str__(self):
        return self._field


def need_to_reshape_array(array, field_size):
    """Check to see if an array needs to be resized before storing.

    When possible, a reference to an array is stored. However, if the
    array is not of the correct shape, a view of the array (with
    the correct shape) is stored.

    Parameters
    ----------
    array : numpy array
        Numpy array to check.
    field_size : int
        Size of the field the array will be placed into.

    Returns
    -------
    bool
        True is the array should be resized.
    """
    if field_size > 1:
        stored_shape = (field_size, )
    else:
        stored_shape = array.squeeze().shape

    return array.shape != stored_shape


class ScalarDataFields(dict):

    """Collection of named data fields that are of the same size.

    Holds a collection of data fields that all contain the same number of
    elements and index each of them with a name. This class inherits from
    a standard Python `dict`, which allows access to the fields through
    dict-like syntax.

    The syntax `.at_[element]` can also be used as syntactic sugar to access
    fields. e.g., `n1 = fields.at_node['name1']`, `n2 = grid.at_link['name2']`.

    Parameters
    ----------
    size : int
        The number of elements in each of the data fields.

    Attributes
    ----------
    units
    size

    See Also
    --------
    landlab.field.ModelDataFields.ones : Hold collections of
        `ScalarDataFields`.

    Examples
    --------
    >>> from landlab.field import ScalarDataFields
    >>> fields = ScalarDataFields(5)
    >>> fields.add_field('land_surface__elevation', [1, 2, 3, 4, 5])
    array([1, 2, 3, 4, 5])
    >>> fields['air__temperature'] = [2, 3, 4, 5, 6]
    >>> fields['land_surface__temperature'] = [0, 1]
    ...     # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ValueError: total size of the new array must be the same as the field

    Fields can also be multidimensional arrays so long as they can be
    resized such that the first dimension is the size of the field.
    The stored field will be resized view of the input array such that
    the size of the first dimension is the size of the field.

    >>> fields['air__temperature'] = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    >>> fields['air__temperature']
    array([[ 2,  3],
           [ 4,  5],
           [ 6,  7],
           [ 8,  9],
           [10, 11]])

    You can also create unsized fields. These fields will not be sized until
    the first field is added to the collection. Once the size is set, all
    fields must be the same size.

    >>> fields = ScalarDataFields()
    >>> fields['land_surface__temperature'] = [0, 1]
    >>> fields['land_surface__temperature']
    array([0, 1])
    >>> fields['air__temperature'] = [2, 3, 4, 5, 6]
    ...     # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ValueError: total size of the new array must be the same as the field

    Fields defined on a grid, which inherits from the ScalarModelFields class,
    behave similarly, though the length of those fields will be forced
    by the element type they are defined on:

    >>> from landlab import RasterModelGrid
    >>> import numpy as np
    >>> mg = RasterModelGrid((4, 5))
    >>> z = mg.add_field('cell', 'topographic__elevation', np.random.rand(
    ...         mg.number_of_cells), units='m')
    >>> mg.at_cell['topographic__elevation'].size == mg.number_of_cells
    True

    LLCATS: FIELDCR, FIELDIO
    """

    def __init__(self, size=None):
        self._size = size

        super(ScalarDataFields, self).__init__()
        self._units = dict()

    @property
    def units(self):
        """Get units for values of the field.

        Returns
        -------
        str
            Units of the field.
        """
        return self._units

    @property
    def size(self):
        """Number of elements in the field.

        Returns
        -------
        int
            The number of elements in the field.
        """
        return self._size

    @size.setter
    def size(self, size):
        if self._size is None:
            self._size = size
        else:
            raise ValueError('size has already been set')

    def empty(self, **kwds):
        """Uninitialized array whose size is that of the field.

        Return a new array of the data field size, without initializing
        entries. Keyword arguments are the same as that for the equivalent
        numpy function.

        See Also
        --------
        numpy.empty : See for a description of optional keywords.
        landlab.field.ScalarDataFields.ones : Equivalent method that
            initializes the data to 1.
        landlab.field.ScalarDataFields.zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        >>> from landlab.field import ScalarDataFields
        >>> field = ScalarDataFields(4)
        >>> field.empty() # doctest: +SKIP
        array([  2.31584178e+077,  -2.68156175e+154,   9.88131292e-324,
        ... 2.78134232e-309]) # Uninitialized memory

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys())
        []
        """
        return np.empty(self.size, **kwds)

    def ones(self, **kwds):
        """Array, initialized to 1, whose size is that of the field.

        Return a new array of the data field size, filled with ones. Keyword
        arguments are the same as that for the equivalent numpy function.

        See Also
        --------
        numpy.ones : See for a description of optional keywords.
        landlab.field.ScalarDataFields.empty : Equivalent method that
            does not initialize the new array.
        landlab.field.ScalarDataFields.zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        >>> from landlab.field import ScalarDataFields
        >>> field = ScalarDataFields(4)
        >>> field.ones()
        array([ 1.,  1.,  1.,  1.])
        >>> field.ones(dtype=int)
        array([1, 1, 1, 1])

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys())
        []
        """
        return np.ones(self.size, **kwds)

    def zeros(self, **kwds):
        """Array, initialized to 0, whose size is that of the field.

        Return a new array of the data field size, filled with zeros. Keyword
        arguments are the same as that for the equivalent numpy function.

        See Also
        --------
        numpy.zeros : See for a description of optional keywords.
        landlab.field.ScalarDataFields.empty : Equivalent method that does not
            initialize the new array.
        landlab.field.scalar_data_fields.ScalarDataFields.ones : Equivalent
            method that initializes the data to 1.

        Examples
        --------
        >>> from landlab.field import ScalarDataFields
        >>> field = ScalarDataFields(4)
        >>> field.zeros()
        array([ 0.,  0.,  0.,  0.])

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys())
        []
        """
        return np.zeros(self.size, **kwds)

    def add_empty(self, name, units=_UNKNOWN_UNITS, noclobber=True, **kwds):
        """Create and add an uninitialized array of values to the field.

        Create a new array of the data field size, without initializing
        entries, and add it to the field as *name*. The *units* keyword gives
        the units of the new fields as a string. Remaining keyword arguments
        are the same as that for the equivalent numpy function.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        units : str, optional
            Optionally specify the units of the field.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See Also
        --------
        numpy.empty : See for a description of optional keywords.
        landlab.field.ScalarDataFields.empty : Equivalent method that
            does not initialize the new array.
        landlab.field.ScalarDataFields.zeros : Equivalent method that
            initializes the data to 0.

        LLCATS: FIELDCR
        """
        return self.add_field(name, self.empty(**kwds), units=units,
                              noclobber=noclobber)

    def add_ones(self, name, units=_UNKNOWN_UNITS, noclobber=True, **kwds):
        """Create and add an array of values, initialized to 1, to the field.

        Create a new array of the data field size, filled with ones, and
        add it to the field as *name*. The *units* keyword gives the units of
        the new fields as a string. Remaining keyword arguments are the same
        as that for the equivalent numpy function.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        units : str, optional
            Optionally specify the units of the field.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See Also
        --------
        numpy.ones : See for a description of optional keywords.
        andlab.field.ScalarDataFields.add_empty : Equivalent method that
            does not initialize the new array.
        andlab.field.ScalarDataFields.add_zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        Add a new, named field to a collection of fields.

        >>> from landlab.field import ScalarDataFields
        >>> field = ScalarDataFields(4)
        >>> field.add_ones('topographic__elevation')
        array([ 1.,  1.,  1.,  1.])
        >>> list(field.keys())
        ['topographic__elevation']
        >>> field['topographic__elevation']
        array([ 1.,  1.,  1.,  1.])

        LLCATS: FIELDCR
        """
        return self.add_field(name, self.ones(**kwds), units=units,
                              noclobber=noclobber)

    def add_zeros(self, name, units=_UNKNOWN_UNITS, noclobber=True, **kwds):
        """Create and add an array of values, initialized to 0, to the field.

        Create a new array of the data field size, filled with zeros, and
        add it to the field as *name*. The *units* keyword gives the units of
        the new fields as a string. Remaining keyword arguments are the same
        as that for the equivalent numpy function.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        units : str, optional
            Optionally specify the units of the field.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See also
        --------
        numpy.zeros : See for a description of optional keywords.
        landlab.field.ScalarDataFields.add_empty : Equivalent method that
            does not initialize the new array.
        landlab.field.ScalarDataFields.add_ones : Equivalent method that
            initializes the data to 1.

        LLCATS: FIELDCR
        """
        return self.add_field(name, self.zeros(**kwds), units=units,
                              noclobber=noclobber)

    def add_field(self, name, value_array, units=_UNKNOWN_UNITS, copy=False,
                  noclobber=True, **kwds):
        """Add an array of values to the field.

        Add an array of data values to a collection of fields and associate it
        with the key, *name*. Use the *copy* keyword to, optionally, add a
        copy of the provided array.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        value_array : numpy.array
            Array of values to add to the field.
        units : str, optional
            Optionally specify the units of the field.
        copy : boolean, optional
            If True, add a *copy* of the array to the field. Otherwise save add
            a reference to the array.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        numpy.array
            The data array added to the field. Depending on the *copy*
            keyword, this could be a copy of *value_array* or *value_array*
            itself.

        Raises
        ------
        ValueError :
            If *value_array* has a size different from the field.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.field import ScalarDataFields
        >>> field = ScalarDataFields(4)
        >>> values = np.ones(4, dtype=int)
        >>> field.add_field('topographic__elevation', values)
        array([1, 1, 1, 1])

        A new field is added to the collection of fields. The saved value
        array is the same as the one initially created.

        >>> field['topographic__elevation'] is values
        True

        If you want to save a copy of the array, use the *copy* keyword. In
        addition, adding values to an existing field will remove the reference
        to the previously saved array. The *noclobber* keyword changes this
        behavior to raise an exception in such a case.

        >>> field.add_field('topographic__elevation', values, copy=True,
        ...                 noclobber=False)
        array([1, 1, 1, 1])
        >>> field['topographic__elevation'] is values
        False
        >>> field.add_field('topographic__elevation', values, noclobber=True)
        ...     # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        FieldError: topographic__elevation already exists

        LLCATS: FIELDCR
        """
        if noclobber and name in self:
            raise FieldError('{name}: already exists'. format(name=name))

        value_array = np.asarray(value_array)

        if copy:
            value_array = value_array.copy()

        self[name] = value_array

        self.set_units(name, units)
        return self[name]

    def set_units(self, name, units):
        """Set the units for a field of values.

        Parameters
        ----------
        name: str
            Name of the field.
        units: str
            Units for the field

        Raises
        ------
        KeyError
            If the named field does not exist.

        LLCATS: FIELDCR
        """
        self._units[name] = units

    def __setitem__(self, name, value_array):
        """Store a data field by name."""
        value_array = np.asarray(value_array)

        if self.size is None:
            self.size = value_array.size

        if need_to_reshape_array(value_array, self.size):
            value_array = value_array.reshape((self.size, -1)).squeeze()

        if name not in self:
            self.set_units(name, None)

        super(ScalarDataFields, self).__setitem__(name, value_array)

    def __getitem__(self, name):
        """Get a data field by name."""
        try:
            return super(ScalarDataFields, self).__getitem__(name)
        except KeyError:
            raise FieldError(name)




import numpy as np
cimport numpy as np
cimport cython

cdef extern from "math.h":
    double exp(double x) nogil

DTYPE_FLOAT = np.double
ctypedef np.double_t DTYPE_FLOAT_t

DTYPE_INT = np.int
ctypedef np.int_t DTYPE_INT_t


def calculate_qs_in(np.ndarray[DTYPE_INT_t, ndim=1] stack_flip_ud,
                    np.ndarray[DTYPE_INT_t, ndim=1] flow_receivers,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] cell_area_at_node,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] q,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] qs,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] qs_in,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] Es,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] Er,
                    DTYPE_FLOAT_t v_s,
                    DTYPE_FLOAT_t F_f,
                    DTYPE_FLOAT_t phi):
    """Calculate and qs and qs_in."""
    # define internal variables
    cdef unsigned int n_nodes = stack_flip_ud.size
    cdef unsigned int node_id
    cdef unsigned int i

    # iterate top to bottom through the stack, calculate qs and adjust  qs_in
    for i in range(n_nodes):

        # choose the node id
        node_id = stack_flip_ud[i]

        # If q at current node is greather than zero, calculate qs based on a
        # local analytical solution. This local analytical solution depends on
        # qs_in, the sediment flux coming into the node from upstream (hence
        # the upstream to downstream node ordering).

        # Because calculation of qs requires qs_in, this operation must be done
        # in an upstream to downstream loop, and cannot be vectorized.

        if q[node_id] > 0:
            qs[node_id] = (qs_in[node_id] + (((1. - phi) * Es[node_id]) + ((1.0 - F_f) * (Er[node_id]))) * cell_area_at_node[node_id]) / \
                            (1.0 + (v_s * cell_area_at_node[node_id] / (q[node_id])))

            # finally, add this nodes qs to recieiving nodes qs_in.
            # if qs[node_id] == 0, then there is no need for this line to be
            # evaluated.
            qs_in[flow_receivers[node_id]] += qs[node_id]

        else:
            # if q at the current node is zero, set qs at that node is zero.
            qs[node_id] = 0



# -*- coding: utf-8 -*-
"""
test_lake_mapper:

Created on Sun Sep 27 09:52:50, 2015

@author: gtucker, amended dejh
"""
from pytest import approx

import landlab
from landlab import RasterModelGrid
from landlab.components.flow_routing import FlowRouter, DepressionFinderAndRouter
from numpy import sin, pi
import numpy as np  # for use of np.round
from numpy.testing import assert_array_equal
from landlab import BAD_INDEX_VALUE as XX


NUM_GRID_ROWS = 8
NUM_GRID_COLS = 8
PERIOD_X = 8.
PERIOD_Y = 4.


def create_test_grid():
    """
    Create a test grid and elevation field with sinusoidal depressions and
    hills.
    """
    # Create grid
    rmg = RasterModelGrid(NUM_GRID_ROWS, NUM_GRID_COLS)

    # Create topography field
    z = rmg.add_zeros("node", "topographic__elevation")

    # Make topography into sinusoidal hills and depressions
    z[:] = sin(2 * pi * rmg.node_x / PERIOD_X) * sin(2 * pi * rmg.node_y / PERIOD_Y)

    # Set 3 sides of the grid to be closed boundaries
    rmg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    return rmg


def check_fields1(grid):
    """
    Check to make sure the right fields have been created.
    """
    try:
        grid.at_node["topographic__elevation"]
        grid.at_node["flood_status_code"]
        grid.at_node["depression__depth"]
        grid.at_node["depression__outlet_node"]
        grid.at_node["is_pit"]
    except:
        print("Test failure in check_fields")
        raise


def check_array_values1(rmg, lm):
    """
    Check values of the various fields against known values.
    """

    assert_array_equal(
        lm.is_pit,
        [
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            True,
            False,
            False,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ],
    )

    assert_array_equal(
        lm.flood_status,
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            3,
            3,
            0,
            0,
            0,
            0,
            0,
            0,
            3,
            3,
            0,
            0,
            3,
            3,
            3,
            3,
            0,
            0,
            0,
            0,
            3,
            3,
            3,
            3,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            3,
            3,
            0,
            0,
            3,
            0,
            0,
            0,
            3,
            3,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
    )

    dd1 = np.round(lm.depression_depth * 100)
    assert_array_equal(
        dd1,
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            71.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ],
    )

    dd1 = np.round(rmg.at_node["depression__depth"] * 100)
    assert_array_equal(
        dd1,
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            71.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ],
    )

    assert_array_equal(
        lm.depression_outlet_map,
        [
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            5,
            5,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            50,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
        ],
    )

    assert_array_equal(
        rmg.at_node["depression__outlet_node"],
        [
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            5,
            5,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            50,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
        ],
    )


def check_fields2(grid):
    """
    Check to make sure the right fields have been created.
    """
    try:
        grid.at_node["topographic__elevation"]
        grid.at_node["flood_status_code"]
        grid.at_node["depression__depth"]
        grid.at_node["depression__outlet_node"]
        grid.at_node["is_pit"]
    except:
        print("Test failure in check_fields")
        raise


def check_array_values2(rmg, lm):
    """
    Check values of the various fields against known values.
    """
    F = False
    T = True
    assert_array_equal(
        lm.is_pit,
        [
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            T,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            T,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            T,
            F,
            F,
            T,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
            F,
        ],
    )

    assert_array_equal(
        lm.flood_status,
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            3,
            3,
            0,
            0,
            0,
            0,
            0,
            0,
            3,
            3,
            0,
            0,
            3,
            3,
            3,
            3,
            0,
            0,
            0,
            0,
            3,
            3,
            3,
            3,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            3,
            3,
            0,
            0,
            3,
            0,
            0,
            0,
            3,
            3,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
    )

    dd1 = np.round(lm.depression_depth * 100)
    assert_array_equal(
        dd1,
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            71.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ],
    )

    dd1 = np.round(rmg.at_node["depression__depth"] * 100)
    assert_array_equal(
        dd1,
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            71.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            71.,
            100.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ],
    )

    assert_array_equal(
        lm.depression_outlet_map,
        [
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            5,
            5,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            50,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
        ],
    )

    assert_array_equal(
        rmg.at_node["depression__outlet_node"],
        [
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            5,
            5,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            50,
            XX,
            XX,
            XX,
            5,
            5,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
        ],
    )


def test_lake_mapper():
    """
    Create a test grid and run a series of tests.
    """
    # Make a test grid
    rmg = create_test_grid()

    # Instantiate a lake mapper
    # (Note that we don't need to send it an input file name, because our grid
    # already has a topographic__elevation field)
    lm = DepressionFinderAndRouter(rmg)

    # Run it on our test grid
    lm.map_depressions()

    # Run tests
    check_fields1(rmg)
    check_array_values1(rmg, lm)
    check_fields2(rmg)
    check_array_values2(rmg, lm)


def test_initial_routing(dans_grid3):
    """
    Test the action of fr.route_flow() on the grid.
    """
    dans_grid3.fr.route_flow()
    assert dans_grid3.mg.at_node["flow__receiver_node"] == approx(dans_grid3.r_old)
    assert dans_grid3.mg.at_node["drainage_area"] == approx(dans_grid3.A_old)


def test_rerouting_with_supplied_pits(dans_grid3):
    """
    Test with the output from a successful run of fr.route_flow.
    """
    dans_grid3.fr.route_flow()
    assert_array_equal(
        dans_grid3.mg.at_node["flow__link_to_receiver_node"], dans_grid3.links_old
    )
    dans_grid3.lf.map_depressions()
    assert_array_equal(dans_grid3.mg.at_node["flow__receiver_node"], dans_grid3.r_new)
    assert dans_grid3.mg.at_node["drainage_area"] == approx(dans_grid3.A_new)
    assert dans_grid3.mg.at_node["surface_water__discharge"] == approx(dans_grid3.A_new)
    assert dans_grid3.mg.at_node["flow__upstream_node_order"] == approx(
        dans_grid3.s_new
    )
    assert dans_grid3.mg.at_node["flow__link_to_receiver_node"] == approx(
        dans_grid3.links_new
    )


def test_changing_slopes(dans_grid3):
    """
    Test with the output from a successful run of fr.route_flow.
    """
    slope_old = np.array(
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            2.,
            2.,
            2.,
            2.,
            2.,
            0.,
            0.,
            2.,
            0.1,
            0.,
            0.1,
            2.,
            0.,
            0.,
            2.,
            0.14142136,
            0.1,
            0.14142136,
            2.,
            0.,
            0.,
            2.,
            1.2,
            1.,
            1.,
            2.,
            0.,
            0.,
            1.06066017,
            1.1,
            1.06066017,
            1.,
            1.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ]
    )
    slope_new = np.array(
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            2.,
            2.,
            2.,
            2.,
            2.,
            0.,
            0.,
            2.,
            0.,
            0.,
            0.,
            2.,
            0.,
            0.,
            2.,
            0.,
            0.,
            0.1,
            2.,
            0.,
            0.,
            2.,
            1.2,
            1.,
            1.,
            2.,
            0.,
            0.,
            1.06066017,
            1.1,
            1.06066017,
            1.,
            1.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ]
    )
    dans_grid3.fr.run_one_step()
    assert dans_grid3.mg.at_node["topographic__steepest_slope"] == approx(slope_old)
    dans_grid3.lf.map_depressions()
    assert dans_grid3.mg.at_node["topographic__steepest_slope"] == approx(slope_new)


def test_filling_alone(dans_grid3):
    """
    Test the filler alone, w/o supplying information on the pits.
    
    Setting the the *pits* parameter to None causes the mapper to look for pits
    using its _find_pits method.
    """
    dans_grid3.lf.map_depressions(pits=None, reroute_flow=False)
    assert_array_equal(
        dans_grid3.mg.at_node["flow__receiver_node"], XX * np.ones(49, dtype=int)
    )
    assert_array_equal(
        dans_grid3.lf.depression_outlet_map, dans_grid3.depr_outlet_target
    )


def test_filling_supplied_pits(dans_grid3):
    """
    Test the filler without rereouting, but confusingly, where there *is*
    aready routing information available!
    Also tests the supply of an array for 'pits'
    """
    dans_grid3.fr.route_flow()
    dans_grid3.lf.map_depressions(
        pits=dans_grid3.mg.at_node["flow__sink_flag"], reroute_flow=False
    )
    assert_array_equal(dans_grid3.mg.at_node["flow__receiver_node"], dans_grid3.r_old)


def test_pits_as_IDs(dans_grid3):
    """
    Smoke test for passing specific IDs, not an array, to the mapper.
    """
    dans_grid3.fr.route_flow()
    dans_grid3.lf.map_depressions(
        pits=np.where(dans_grid3.mg.at_node["flow__sink_flag"])[0]
    )
    assert dans_grid3.mg.at_node["drainage_area"] == approx(dans_grid3.A_new)


def test_edge_draining():
    """
    This tests when the lake attempts to drain from an edge, where an issue
    is suspected.
    """
    # Create a 7x7 test grid with a well defined hole in it, AT THE EDGE.
    mg = RasterModelGrid((7, 7), (1., 1.))

    z = mg.node_x.copy()
    guard_sides = np.concatenate((np.arange(7, 14), np.arange(35, 42)))
    edges = np.concatenate((np.arange(7), np.arange(42, 49)))
    hole_here = np.array(([15, 16, 22, 23, 29, 30]))
    z[guard_sides] = z[13]
    z[edges] = -2.  # force flow outwards from the tops of the guards
    z[hole_here] = -1.

    A_new = np.array(
        [
            [
                [
                    0.,
                    1.,
                    1.,
                    1.,
                    1.,
                    1.,
                    0.,
                    0.,
                    1.,
                    1.,
                    1.,
                    1.,
                    1.,
                    0.,
                    15.,
                    5.,
                    4.,
                    3.,
                    2.,
                    1.,
                    0.,
                    0.,
                    10.,
                    4.,
                    3.,
                    2.,
                    1.,
                    0.,
                    0.,
                    1.,
                    4.,
                    3.,
                    2.,
                    1.,
                    0.,
                    0.,
                    1.,
                    1.,
                    1.,
                    1.,
                    1.,
                    0.,
                    0.,
                    1.,
                    1.,
                    1.,
                    1.,
                    1.,
                    0.,
                ]
            ]
        ]
    ).flatten()

    depr_outlet_target = np.array(
        [
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            14,
            14,
            XX,
            XX,
            XX,
            XX,
            XX,
            14,
            14,
            XX,
            XX,
            XX,
            XX,
            XX,
            14,
            14,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
        ]
    ).flatten()

    mg.add_field("node", "topographic__elevation", z, units="-")

    fr = FlowRouter(mg)
    lf = DepressionFinderAndRouter(mg)

    fr.route_flow()
    lf.map_depressions()
    assert mg.at_node["drainage_area"] == approx(A_new)
    assert lf.depression_outlet_map == approx(depr_outlet_target)


def test_degenerate_drainage():
    """
    This "hourglass" configuration should be one of the hardest to correctly
    re-route.
    """
    mg = RasterModelGrid(9, 5)
    z_init = mg.node_x.copy() * 0.0001 + 1.
    lake_pits = np.array([7, 11, 12, 13, 17, 27, 31, 32, 33, 37])
    z_init[lake_pits] = -1.
    z_init[22] = 0.  # the common spill pt for both lakes
    z_init[21] = 0.1  # an adverse bump in the spillway
    z_init[20] = -0.2  # the spillway
    z = mg.add_field("node", "topographic__elevation", z_init)

    fr = FlowRouter(mg)
    lf = DepressionFinderAndRouter(mg)
    fr.route_flow()
    lf.map_depressions()

    #    correct_A = np.array([ 0.,   0.,   0.,   0.,   0.,
    #                           0.,   1.,   3.,   1.,   0.,
    #                           0.,   5.,   1.,   2.,   0.,
    #                           0.,   1.,  10.,   1.,   0.,
    #                          21.,  21.,   1.,   1.,   0.,
    #                           0.,   1.,   9.,   1.,   0.,
    #                           0.,   3.,   1.,   2.,   0.,
    #                           0.,   1.,   1.,   1.,   0.,
    #                           0.,   0.,   0.,   0.,   0.])

    correct_A = np.array(
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            1.,
            3.,
            1.,
            0.,
            0.,
            2.,
            4.,
            2.,
            0.,
            0.,
            1.,
            10.,
            1.,
            0.,
            21.,
            21.,
            1.,
            1.,
            0.,
            0.,
            1.,
            9.,
            1.,
            0.,
            0.,
            2.,
            2.,
            2.,
            0.,
            0.,
            1.,
            1.,
            1.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ]
    )

    thelake = np.concatenate((lake_pits, [22])).sort()

    assert mg.at_node["drainage_area"] == approx(correct_A)


def test_three_pits():
    """
    A test to ensure the component correctly handles cases where there are
    multiple pits.
    """
    mg = RasterModelGrid(10, 10, 1.)
    z = mg.add_field("node", "topographic__elevation", mg.node_x.copy())
    # a sloping plane
    # np.random.seed(seed=0)
    # z += np.random.rand(100)/10000.
    # punch some holes
    z[33] = 1.
    z[43] = 1.
    z[37] = 4.
    z[74:76] = 1.
    fr = FlowRouter(mg)
    lf = DepressionFinderAndRouter(mg)
    fr.route_flow()
    lf.map_depressions()

    flow_sinks_target = np.zeros(100, dtype=bool)
    flow_sinks_target[mg.boundary_nodes] = True
    # no internal sinks now:
    assert_array_equal(mg.at_node["flow__sink_flag"], flow_sinks_target)

    # test conservation of mass:
    assert mg.at_node["drainage_area"].reshape((10, 10))[1:-1, 1].sum() == approx(
        8. ** 2
    )
    # ^all the core nodes

    # test the actual flow field:
    nA = np.array(
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            8.,
            8.,
            7.,
            6.,
            5.,
            4.,
            3.,
            2.,
            1.,
            0.,
            2.,
            2.,
            1.,
            1.,
            2.,
            1.,
            1.,
            1.,
            1.,
            0.,
            26.,
            26.,
            25.,
            15.,
            11.,
            10.,
            9.,
            8.,
            1.,
            0.,
            2.,
            2.,
            1.,
            9.,
            2.,
            1.,
            1.,
            1.,
            1.,
            0.,
            2.,
            2.,
            1.,
            1.,
            5.,
            4.,
            3.,
            2.,
            1.,
            0.,
            2.,
            2.,
            1.,
            1.,
            1.,
            1.,
            3.,
            2.,
            1.,
            0.,
            20.,
            20.,
            19.,
            18.,
            17.,
            12.,
            3.,
            2.,
            1.,
            0.,
            2.,
            2.,
            1.,
            1.,
            1.,
            1.,
            3.,
            2.,
            1.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ]
    )
    assert_array_equal(mg.at_node["drainage_area"], nA)

    # test a couple more properties:
    lc = np.empty(100, dtype=int)
    lc.fill(XX)
    lc[33] = 33
    lc[43] = 33
    lc[37] = 37
    lc[74:76] = 74
    assert_array_equal(lf.lake_map, lc)
    assert_array_equal(lf.lake_codes, [33, 37, 74])
    assert lf.number_of_lakes == 3
    assert lf.lake_areas == approx([2., 1., 2.])
    assert lf.lake_volumes == approx([2., 2., 4.])


def test_composite_pits():
    """
    A test to ensure the component correctly handles cases where there are
    multiple pits, inset into each other.
    """
    mg = RasterModelGrid(10, 10, 1.)
    z = mg.add_field("node", "topographic__elevation", mg.node_x.copy())
    # a sloping plane
    # np.random.seed(seed=0)
    # z += np.random.rand(100)/10000.
    # punch one big hole
    z.reshape((10, 10))[3:8, 3:8] = 0.
    # dig a couple of inset holes
    z[57] = -1.
    z[44] = -2.
    z[54] = -10.

    # make an outlet
    z[71] = 0.9

    fr = FlowRouter(mg)
    lf = DepressionFinderAndRouter(mg)
    fr.route_flow()
    lf.map_depressions()

    flow_sinks_target = np.zeros(100, dtype=bool)
    flow_sinks_target[mg.boundary_nodes] = True
    # no internal sinks now:
    assert_array_equal(mg.at_node["flow__sink_flag"], flow_sinks_target)

    # test conservation of mass:
    assert mg.at_node["drainage_area"].reshape((10, 10))[1:-1, 1].sum() == approx(
        8. ** 2
    )
    # ^all the core nodes

    # test the actual flow field:
    #    nA = np.array([  0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,
    #                     8.,   8.,   7.,   6.,   5.,   4.,   3.,   2.,   1.,   0.,
    #                     1.,   1.,   1.,   1.,   1.,   1.,   1.,   1.,   1.,   0.,
    #                     1.,   1.,   1.,   4.,   2.,   2.,   8.,   4.,   1.,   0.,
    #                     1.,   1.,   1.,   8.,   3.,  15.,   3.,   2.,   1.,   0.,
    #                     1.,   1.,   1.,  13.,  25.,   6.,   3.,   2.,   1.,   0.,
    #                     1.,   1.,   1.,  45.,   3.,   3.,   5.,   2.,   1.,   0.,
    #                    50.,  50.,  49.,   3.,   2.,   2.,   2.,   4.,   1.,   0.,
    #                     1.,   1.,   1.,   1.,   1.,   1.,   1.,   1.,   1.,   0.,
    #                     0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.])
    nA = np.array(
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            8.,
            8.,
            7.,
            6.,
            5.,
            4.,
            3.,
            2.,
            1.,
            0.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            0.,
            1.,
            1.,
            1.,
            4.,
            2.,
            2.,
            6.,
            4.,
            1.,
            0.,
            1.,
            1.,
            1.,
            6.,
            3.,
            12.,
            3.,
            2.,
            1.,
            0.,
            1.,
            1.,
            1.,
            8.,
            20.,
            4.,
            3.,
            2.,
            1.,
            0.,
            1.,
            1.,
            1.,
            35.,
            5.,
            4.,
            3.,
            2.,
            1.,
            0.,
            50.,
            50.,
            49.,
            13.,
            10.,
            8.,
            6.,
            4.,
            1.,
            0.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ]
    )
    assert_array_equal(mg.at_node["drainage_area"], nA)

    # the lake code map:
    lc = np.array(
        [
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            57,
            57,
            57,
            57,
            57,
            XX,
            XX,
            XX,
            XX,
            XX,
            57,
            57,
            57,
            57,
            57,
            XX,
            XX,
            XX,
            XX,
            XX,
            57,
            57,
            57,
            57,
            57,
            XX,
            XX,
            XX,
            XX,
            XX,
            57,
            57,
            57,
            57,
            57,
            XX,
            XX,
            XX,
            XX,
            XX,
            57,
            57,
            57,
            57,
            57,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
            XX,
        ]
    )

    # test the remaining properties:
    assert lf.lake_outlets.size == 1
    assert lf.lake_outlets[0] == 72
    outlets_in_map = np.unique(lf.depression_outlet_map)
    assert outlets_in_map.size == 2
    assert outlets_in_map[1] == 72
    assert lf.number_of_lakes == 1
    assert lf.lake_codes[0] == 57
    assert_array_equal(lf.lake_map, lc)
    assert lf.lake_areas[0] == approx(25.)
    assert lf.lake_volumes[0] == approx(63.)


def test_D8_D4_fill(d4_grid):
    """
    Tests the functionality of D4 filling.
    """
    d4_grid.lfD8.map_depressions(pits=None, reroute_flow=False)
    d4_grid.lfD4.map_depressions(pits=None, reroute_flow=False)
    assert d4_grid.lfD8.number_of_lakes == 1
    assert d4_grid.lfD4.number_of_lakes == 3

    correct_D8_lake_map = np.empty(7 * 7, dtype=int)
    correct_D8_lake_map.fill(XX)
    correct_D8_lake_map[d4_grid.lake_nodes] = 10
    correct_D4_lake_map = correct_D8_lake_map.copy()
    correct_D4_lake_map[d4_grid.lake_nodes[5:]] = 32
    correct_D4_lake_map[d4_grid.lake_nodes[-2]] = 38
    correct_D8_depths = np.zeros(7 * 7, dtype=float)
    correct_D8_depths[d4_grid.lake_nodes] = 2.
    correct_D4_depths = correct_D8_depths.copy()
    correct_D4_depths[d4_grid.lake_nodes[5:]] = 4.
    correct_D4_depths[d4_grid.lake_nodes[-2]] = 3.

    assert_array_equal(d4_grid.lfD8.lake_map, correct_D8_lake_map)
    assert_array_equal(d4_grid.lfD4.lake_map, correct_D4_lake_map)

    assert d4_grid.mg1.at_node["depression__depth"] == approx(correct_D8_depths)
    assert d4_grid.mg2.at_node["depression__depth"] == approx(correct_D4_depths)


def test_D8_D4_route(d4_grid):
    """
    Tests the functionality of D4 routing.
    """
    d4_grid.frD8.route_flow()
    d4_grid.frD4.route_flow()
    d4_grid.lfD8.map_depressions()
    d4_grid.lfD4.map_depressions()
    assert d4_grid.lfD8.number_of_lakes == 1
    assert d4_grid.lfD4.number_of_lakes == 3

    #    flow_recD8 = np.array([ 0,  1,  2,  3,  4,  5,  6,  7, 16, 10, 16, 10, 18,
    #                           13, 14, 14, 15, 16, 10, 18, 20, 21, 16, 16, 16, 18,
    #                           33, 27, 28, 28, 24, 24, 24, 32, 34, 35, 35, 38, 32,
    #                           32, 32, 41, 42, 43, 44, 45, 46, 47, 48])
    flow_recD8 = np.array(
        [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            16,
            10,
            16,
            10,
            18,
            13,
            14,
            14,
            15,
            16,
            17,
            18,
            20,
            21,
            16,
            16,
            16,
            18,
            33,
            27,
            28,
            28,
            24,
            24,
            24,
            32,
            34,
            35,
            35,
            38,
            32,
            32,
            32,
            41,
            42,
            43,
            44,
            45,
            46,
            47,
            48,
        ]
    )
    #    flow_recD4 = np.array([ 0,  1,  2,  3,  4,  5,  6,  7,  7, 10, 17, 10, 11,
    #                           13, 14, 14, 15, 16, 17, 18, 20, 21, 21, 16, 17, 18,
    #                           33, 27, 28, 28, 29, 24, 31, 32, 34, 35, 35, 36, 37,
    #                           32, 33, 41, 42, 43, 44, 45, 46, 47, 48])
    flow_recD4 = np.array(
        [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            7,
            10,
            17,
            10,
            11,
            13,
            14,
            14,
            15,
            16,
            17,
            18,
            20,
            21,
            21,
            16,
            17,
            18,
            33,
            27,
            28,
            28,
            29,
            38,
            31,
            32,
            34,
            35,
            35,
            36,
            37,
            32,
            33,
            41,
            42,
            43,
            44,
            45,
            46,
            47,
            48,
        ]
    )
    assert_array_equal(d4_grid.mg1.at_node["flow__receiver_node"], flow_recD8)
    assert_array_equal(d4_grid.mg2.at_node["flow__receiver_node"], flow_recD4)
    assert d4_grid.mg1.at_node["drainage_area"].reshape((7, 7))[:, 0].sum() == approx(
        d4_grid.mg2.at_node["drainage_area"].reshape((7, 7))[:, 0].sum()
    )



#! /usr/bin/env python
"""
Collections of components
"""

import inspect

#from landlab.interfaces import BmiBase
from .interfaces import BmiBase
from .decorators import is_implementation


class Error(Exception):
    """
    Base exception for this module
    """


class UnknownComponentError(Error):
    """
    Raise this exception when requesting a component unknown to the framework
    """

    def __init__(self, name):
        self._name = name

    def __str__(self):
        return self._name


class MultipleProvidersError(Error):
    """
    Raise this exception if multiple components provide the same variable
    """

    def __init__(self, var_name):
        self._name = var_name

    def __str__(self):
        return 'Multiple component provide variable %s' % self._name


class NoProvidersError(Error):
    """
    Raise this exception if no components provide the a variable
    """

    def __init__(self, var_name):
        self._name = var_name

    def __str__(self):
        return 'No components provide variable %s' % self._name


class BadVarNameError(Error):
    """
    Raise this exception if a variable name is not found for a component.
    """

    def __init__(self, var_name):
        self._name = var_name

    def __str__(self):
        return 'Component does not contain variable %s' % self._name


def get_var_names(component, intent='input'):
    """
    Get a list of input or output variable names from *component* (a BMI-like
    object). Use the *intent* keyword to specify whether to return input or
    output variable names. *intent* must be one of *input* or *output*.
    """
    assert(intent in ['input', 'output'])

    func = getattr(component, 'get_' + intent + '_var_names')
    try:
        var_names = func()
    except TypeError:
        var_names = getattr(component, '_' + intent + '_var_names')

    return var_names


class Collection(dict):
    """
    A collection of components that implement a BmiBase
    """

    def __init__(self, **kwds):
        super(Collection, self).__init__(**kwds)
        for (_, component) in self.items():
            assert(is_implementation(component, BmiBase))

    def list(self):
        """
        Get a list of component names in the collection.
        """
        return self.keys()

    def uses(self):
        """
        Get a list of variable names that components in the collection uses.
        """
        return self._var_names(intent='input')

    def provides(self):
        """
        Get a list of variable names that components in the collection
        provides.
        """
        return self._var_names(intent='output')

    def find_provider(self, var_name):
        """
        Find a component in the collection that provides the CSDMS standard name,
        *var_name*. The returned list contains the names of the components
        providing *var_name*.
        """
        return self._find_var_name(var_name, intent='output')

    def find_user(self, var_name):
        """
        Find components in the collection that use the specified CSDMS standard
        variable name, *var_name*. The returned list contains the names of the
        components using *var_name*.
        """
        return self._find_var_name(var_name, intent='input')

    def find_connections(self):
        """
        Find connections between components that use and provide variables.
        The returned dictionary will have keys that are names of components
        that provide varaibles. Its keys will be a dictionary of variable
        names and list of components that use the variable.
        """
        connections = dict()
        for var_name in self.uses():
            providers = self.find_provider(var_name)
            if len(providers) > 1:
                raise MultipleProvidersError(var_name)
            elif len(providers) == 0:
                raise NoProvidersError(var_name)
            else:
                try:
                    connections[providers[0]].update({
                        var_name: self.find_user(var_name),
                    })
                except KeyError:
                    connections[providers[0]] = {
                        var_name: self.find_user(var_name),
                    }
        return connections

    def _find_var_name(self, needle, intent='input'):
        """
        Find a variable either used or provided by components in the
        collection.

        :needle: The variable name to search for
        :keyword intent: One of 'input', or 'output'

        :returns: List of component names
        """
        assert(intent in ['input', 'output'])

        names = set()
        for (name, component) in self.items():
            try:
                for var_name in get_var_names(component, intent=intent):
                    if var_name == needle:
                        names.add(name)
            except AttributeError:
                pass
        return list(names)

    def _var_names(self, intent='input'):
        """
        Get variable names either used or provided by components in the
        collection.

        :keyword intent: One of 'input', or 'output'

        :returns: List of variable names
        """
        assert(intent in ['input', 'output'])

        names = set()
        for (name, component) in self.items():
            try:
                for var_name in get_var_names(component, intent=intent):
                    names.add(var_name)
            except AttributeError:
                pass
        return list(names)


class Palette(Collection):
    """
    A collection of component classes that have yet to be instantiated.
    """

    def __init__(self, *args, **kwds):
        super(Palette, self).__init__(*args, **kwds)


class Arena(Collection):
    """
    A collection of component instances.
    """

    def __init__(self):
        super(Arena, self).__init__()
        self._connections = dict()

    def instantiate(self, cls, name):
        """
        Instantiate a component and call it *name*. The component, *cls*, is
        either an instance of a BMI-like object or a class that implements the
        BMI. If *cls* is a class, it is instantiated by calling it's __init__
        method without any arguments.
        """
        if inspect.isclass(cls):
            self[name] = cls()
        else:
            self[name] = cls

        try:
            assert(is_implementation(type(self[name]), BmiBase))
        except AssertionError:
            self.pop(name)
            raise TypeError(
                'Class is not an implementation of cmt.bmi.BmiBase')
        else:
            self._connections[name] = dict()

    def connect(self, user_name, provider_name, var_name):
        """
        Connect two components through a variable. *user_name* is the name of the
        component that uses *var_name* and *provider_name* is the name of the
        component that provides *var_name*.

        If the arena doesn't contain either *user_name* or *provider_name*,
        :func:`UnknownComponentError` is raised.
        """
        try:
            user = self[user_name]
        except KeyError:
            raise UnknownComponentError(user_name)
        try:
            provider = self[provider_name]
        except KeyError:
            raise UnknownComponentError(provider_name)

        try:
            user.set_value(var_name, provider.get_value(var_name))
        except BadVarNameError:
            raise
        except AttributeError:
            raise

        self._connections[user_name].update({var_name: provider_name})

    def _component_children(self, name):
        """
        Get a set of component names that represent the components that
        provide data to a component.

        :name: Name of the base component

        :return: A set of component names
        """
        children = set()
        for child in self._connections[name].values():
            children.add(child)
        return children

    def walk(self, root, tree=None):
        """Walk a connected set of components.

        Walk a connected set of components with the component named *root*. If
        the *tree* keyword is given, treat it as a list of components already
        in the tree and add to that list. If a component is already in the tree,
        do not iterate through that component. This function returns a list of
        component names in the order they are visited by the walk.
        """
        if tree is None:
            tree = []
        if root in tree:
            return tree

        tree.append(root)
        for child in self._component_children(root):
            if child in tree:
                continue
            else:
                self.walk(child, tree=tree)
        return tree



#! /usr/env/python
"""
Python implementation of RadialModelGrid, a grid class used to create and
manage structured Voronoi-Delaunay grids for 2D numerical models.

Do NOT add new documentation here. Grid documentation is now built in a semi-
automated fashion. To modify the text seen on the web, edit the files
`docs/text_for_[gridfile].py.txt`.
"""

import numpy
from six.moves import range

from .voronoi import VoronoiDelaunayGrid
from landlab.utils.decorators import deprecated


class RadialModelGrid(VoronoiDelaunayGrid):

    """Grid of concentric circles.

    This inherited class implements a circular grid in which grid nodes are
    placed at regular radial and semi-regular arc-wise intervals. That is,
    if the radial spacing between *shells* is *dr*, the nodes are placed around
    the circular shell at regular intervals that get as close as possible to
    *dr*. The points are then arranged in a Delaunay triangulation with Voronoi
    cells. Within each ring, nodes are numbered according to Landlab
    convention, from the first node counterclockwise of east. Numbering
    begins at the centermost node and works outwards through the rings.

    Parameters
    ----------
    num_shells : int
        Number of rings in the grid.
    dr : float, optional
        Radial interval for rings.
    origin_x : float, optional
        x-coordinate of origin node.
    origin_y : float, optional
        y-coordinate of origin node.

    Returns
    -------
    RadialModelGrid
        A newly-created grid.

    Examples
    --------
    A grid with just one ring will have a node at the origin surrounded
    by six other nodes.

    >>> from landlab import RadialModelGrid
    >>> omg = RadialModelGrid(num_shells=1, dr=1., origin_x=0., origin_y=0.)
    >>> omg.number_of_nodes
    7
    >>> omg.number_of_cells
    1

    A second rings will have 13 nodes.

    >>> omg = RadialModelGrid(2)
    >>> omg.number_of_nodes
    20
    """

    def __init__(self, num_shells=0, dr=1.0, origin_x=0.0, origin_y=0.0,
                 **kwds):
        """Create a circular grid.

        Create a circular grid in which grid nodes are placed at regular
        radial and semi-regular arc-wise intervals. That is, if the radial
        spacing between *shells* is *dr*, the nodes are placed around the
        circular shell at regular intervals that get as close as possible to
        *dr*.  The points are then arranged in a Delaunay triangulation with
        Voronoi cells.

        Parameters
        ----------
        num_shells : int
            Number of rings in the grid.
        dr : float, optional
            Radial interval for rings.
        origin_x : float, optional
            x-coordinate of origin node.
        origin_y : float, optional
            y-coordinate of origin node.

        Returns
        -------
        RadialModelGrid
            A newly-created grid.

        Examples
        --------
        A grid with just one ring will have a node at the origin surrounded
        by six other nodes.

        >>> from landlab import RadialModelGrid
        >>> omg = RadialModelGrid(num_shells=1, dr=1., origin_x=0.,
        ...                       origin_y=0.)
        >>> omg.number_of_nodes
        7
        >>> omg.number_of_cells
        1

        A second rings will have 13 nodes.

        >>> omg = RadialModelGrid(2)
        >>> omg.number_of_nodes
        20
        """
        # Set number of nodes, and initialize if caller has given dimensions
        self._origin_x = origin_x
        self._origin_y = origin_y
        if num_shells > 0:
            self._initialize(num_shells, dr, origin_x, origin_y)
        super(RadialModelGrid, self).__init__(**kwds)

    @classmethod
    def from_dict(cls, params):
        """
        LLCATS: GINF
        """
        num_shells = params['num_shells']
        dr = params.get('dr', 1.)
        origin = params.get('origin', (0., 0.))

        return cls(num_shells=num_shells, dr=dr, origin_x=origin[0],
                   origin_y=origin[1])

    def _initialize(self, num_shells, dr, origin_x=0.0, origin_y=0.0):
        [pts, npts] = self._create_radial_points(num_shells, dr)
        self._n_shells = int(num_shells)
        self._dr = dr
        super(RadialModelGrid, self)._initialize(pts[:, 0], pts[:, 1])

    def _create_radial_points(self, num_shells, dr, origin_x=0.0,
                              origin_y=0.0):
        """Create a set of points on concentric circles.

        Creates and returns a set of (x,y) points placed in a series of
        concentric circles around the origin.
        """
        shells = numpy.arange(0, num_shells) + 1
        twopi = 2 * numpy.pi
        # number of points in each shell
        n_pts_in_shell = numpy.round(twopi * shells)
        dtheta = twopi / n_pts_in_shell
        npts = int(sum(n_pts_in_shell) + 1)
        pts = numpy.zeros((npts, 2))
        r = shells * dr
        startpt = 1
        for i in numpy.arange(0, num_shells):
            theta = (dtheta[i] * numpy.arange(0, n_pts_in_shell[i]) +
                     dtheta[i] / (i + 1))
            ycoord = r[i] * numpy.sin(theta)
            if numpy.isclose(ycoord[-1], 0.):
                # this modification necessary to force the first ring to
                # follow our new CCW from E numbering convention (DEJH, Nov15)
                ycoord[-1] = 0.
                pts[startpt:(startpt + int(n_pts_in_shell[i])),
                    0] = numpy.roll(r[i] * numpy.cos(theta), 1)
                pts[startpt:(startpt + int(n_pts_in_shell[i])),
                    1] = numpy.roll(ycoord, 1)
            else:
                pts[startpt:(startpt + int(n_pts_in_shell[i])),
                    0] = r[i] * numpy.cos(theta)
                pts[startpt:(startpt + int(n_pts_in_shell[i])),
                    1] = ycoord
            startpt += int(n_pts_in_shell[i])
        pts[:, 0] += origin_x
        pts[:, 1] += origin_y

        return pts, npts

    @property
    def number_of_shells(self):
        """Number of node shells in grid.

        Returns
        -------
        int
            The number of node shells in the radial grid (not counting the
            center node).

        LLCATS: GINF
        """
        return self._n_shells

    @property
    @deprecated(use='spacing_of_shells', version=1.0)
    def shell_spacing(self):
        """Fixed distance between shells.

        LLCATS: DEPR GINF MEAS
        """
        return self._dr

    @property
    def spacing_of_shells(self):
        """Fixed distance between shells.

        LLCATS: GINF MEAS
        """
        return self._dr

    @property
    def number_of_nodes_in_shell(self):
        """Number of nodes in each shell.

        Returns
        -------
        int
            Number of nodes in each shell, excluding the center node.

        LLCATS: GINF NINF
        """
        try:
            return self._nnodes_inshell
        except AttributeError:
            n_pts_in_shell = numpy.round(2. * numpy.pi * (
                numpy.arange(self.number_of_shells, dtype=float) + 1.))
            self._nnodes_inshell = n_pts_in_shell.astype(int)
            return self._nnodes_inshell

    @property
    def radius_at_node(self):
        """Distance for center node to each node.

        Returns
        -------
        ndarray of float
            The distance from the center node of each node.

        >>> mg = RadialModelGrid(num_shells=2)
        >>> mg.radius_at_node
        array([ 2.,  2.,  2.,  2.,  2.,  1.,  1.,  2.,  0.,  1.,  1.,  2.,  2.,
                1.,  1.,  2.,  2.,  2.,  2.,  2.])

        LLCATS: NINF MEAS
        """
        try:
            return self._node_radii
        except AttributeError:
            self._node_radii = numpy.sqrt(numpy.square(self.node_x -
                                                       self._origin_x) +
                                          numpy.square(self.node_y -
                                                       self._origin_y))
            return self._node_radii



#!/usr/bin/env python
"""Rock uplift along a normal fault.

Landlab component that implements rock uplift by a normal fault. Note that this
component does not make any attempt to advect topography laterally.

"""

import numpy as np
from landlab import Component, FieldError


TWO_PI = 2.0*np.pi

class NormalFault(Component):
    """NormalFault implements relative rock motion due to a normal fault.

    The fault can have an arbitrary trace given by two points (x1, y1) and
    (x2, y2) in the `fault_trace` input parameter. These value of these points
    is in model-space coordinates and is not based on node id values or number
    of rows and columns.

    This NormalFault component permits two primary methods for enacting fault
    motion.

    1) **run_one_step**: The throw rate is provided through the
    ``fault_throw_rate_through_time`` parameter. This rate can be constant or
    arbitrary. See the NormalFault tutorial in the landlab tutorials repository
    for an extensive example. In this case, the NormalFault component will keep
    track of the cumulative amount of model-run-time and set the rate based on
    interpolating the provided rate-time history. *NOTE: this means that the
    model run timesteps must align with the time-rate relationship provided to
    NormalFault*. Improving this is on the developers todo list but is of low
    priority.

    2) **run_one_earthquake**: A single uplift event of size dz can be specified
    by this method. If NormalFault is used in this way, any specifications
    provided in the ``fault_throw_rate_through_time`` keyword argument will be
    ignored.

    Note that the NormalFault component does not prevent a user from combining
    the **run_one_step** and **run_one_earthquake** methods. It is encumbent
    upon the user, however, to ensure that these two methods are used in
    combination correctly for their specific use case.

    """

    _name = 'NormalFault'

    #_cite_as = """ """

    _input_var_names = (
        'topographic__elevation',
    )

    _output_var_names = (
        'topographic__elevation',
    )

    _var_units = {
        'topographic__elevation': 'm',
    }

    _var_mapping = {
        'topographic__elevation': 'node',
    }

    _var_doc = {
        'topographic__elevation': 'elevation of the ground surface'
    }

    def __init__(self,
                 grid,
                 faulted_surface = 'topographic__elevation',
                 fault_throw_rate_through_time = {'time': [0],
                                                  'rate': [0.001]},
                 fault_dip_angle = 90.0,
                 fault_trace = {'x1': 0,
                                'y1': 0,
                                'x2': 1,
                                'y2': 1},
                include_boundaries = False,
                **kwargs):
        """
        Instantiation of a NormalFault.

        Parameters
        --------
        grid : ModelGrid
        faulted_surface : str or list of str
            Surface that is modified by the NormalFault component. Must be a
            field name or a list of field names if the fault should uplift more
            than one field. Default value is `topographic__elevation`.
            If the faulted surface does not yet exist, it will be ingored. The
            ``run_one_step`` method will check to see an ignored field has been
            added and if it has been, it will modify it.
        fault_throw_rate_through_time : dict, optional
            Dictionary that specifies the time varying throw rate on the fault.
            Expected format is:
            ``fault_throw_rate_through_time = {'time': array, 'rate': array}``
            Default value is a constant rate of 0.001 (units not specified).
            This is acomplished by providing the dictionary
            ``{'time': [0], 'rate': [0.001]}``. NormalFault uses numpy interp
            to interpolate the time and rate pattern to the current model time.
            This function uses the first value for all values less than the
            first, and the last value for all values greater than the last, and
            thus providing only one number results in all times getting a rate
            of that value.
        fault_dip_angle : float, optional
            Dip angle of the fault in degrees.  Default value is 90 degrees.
        fault_trace : dictionary, optional
            Dictionary that specifies the coordinates of two locations on the
            fault trace. Expected format is
            ``fault_trace = {'x1': float, 'y1': float, 'x2': float, 'y2': float}``
            where the vector from ``(x1, y1)`` to ``(x2, y2)`` defines the
            strike of the fault trace. The orientation of the fault dip relative
            to the strike follows the right hand rule.
            Default is for the fault to strike NE.
        include_boundaries : boolean, optional
            Flag to indicate if model grid boundaries should be uplifted. If
            set to ``True`` uplifted model grid boundaries will be set to the
            average value of their upstream nodes. Default value is False.

        Examples
        --------

        Create a grid on which we will run the NormalFault component.

        >>> from landlab import RasterModelGrid
        >>> from landlab.components import NormalFault
        >>> grid = RasterModelGrid((6, 6), spacing=10)

        Add an elevation field.

        >>> z = grid.add_zeros('node', 'topographic__elevation')


        Set the parameter values for the NormalFault component.

        >>> param_dict = {'faulted_surface': 'topographic__elevation',
        ...               'fault_dip_angle': 90.0,
        ...               'fault_throw_rate_through_time': {'time': [0, 9, 10],
        ...                                                 'rate': [0, 0, 0.05]},
        ...               'fault_trace': {'y1': 0,
        ...                                    'x1': 0,
        ...                                    'y2': 30,
        ...                                    'x2': 60},
        ...              'include_boundaries': False}

        Instantiate a NormalFault component.

        >>> nf = NormalFault(grid, **param_dict)
        >>> nf.faulted_nodes.reshape(grid.shape)
        array([[False, False, False, False, False, False],
               [False,  True, False, False, False, False],
               [False,  True,  True,  True, False, False],
               [False,  True,  True,  True,  True, False],
               [False,  True,  True,  True,  True, False],
               [False, False, False, False, False, False]], dtype=bool)

        As we can see, only a subset of the nodes have been identified as
        *faulted nodes*. Because we have set include_boundaries' to False none
        of the boundary nodes are faulted nodes.

        Next we will run the NormalFault for 30 1-year timesteps.

        >>> dt = 1.0
        >>> for i in range(30):
        ...     nf.run_one_step(dt)
        >>> z.reshape(grid.shape)
        array([[ 0.,  0.,  0.,  0.,  0.,  0.],
               [ 0.,  1.,  0.,  0.,  0.,  0.],
               [ 0.,  1.,  1.,  1.,  0.,  0.],
               [ 0.,  1.,  1.,  1.,  1.,  0.],
               [ 0.,  1.,  1.,  1.,  1.,  0.],
               [ 0.,  0.,  0.,  0.,  0.,  0.]])

        This results in uplift of the faulted nodes, as we would expect.

        If the user knows how much uplift (dz) they want to occur in an event,
        they can use the **run_one_earthquake** function with a specified dz.
        In this case fault_throw_rate_through_time will be ignored.

        >>> nf.run_one_earthquake(dz=100)
        >>> z.reshape(grid.shape)
        array([[   0.,    0.,    0.,    0.,    0.,    0.],
               [   0.,  101.,    0.,    0.,    0.,    0.],
               [   0.,  101.,  101.,  101.,    0.,    0.],
               [   0.,  101.,  101.,  101.,  101.,    0.],
               [   0.,  101.,  101.,  101.,  101.,    0.],
               [   0.,    0.,    0.,    0.,    0.,    0.]])

        Next, we make a very simple landscape model. We need a few components
        and we will set include_boundaries to True.

        >>> from landlab.components import FastscapeEroder, FlowAccumulator
        >>> grid = RasterModelGrid((6, 6), spacing=10)
        >>> z = grid.add_zeros('node', 'topographic__elevation')
        >>> param_dict = {'faulted_surface': 'topographic__elevation',
        ...               'fault_dip_angle': 90.0,
        ...               'fault_throw_rate_through_time': {'time': [0, 900, 1000],
        ...                                                 'rate': [0, 0, 0.05]},
        ...               'fault_trace': {'y1': 0,
        ...                                    'x1': 0,
        ...                                    'y2': 30,
        ...                                    'x2': 60},
        ...              'include_boundaries': True}

        >>> nf = NormalFault(grid, **param_dict)
        >>> fr = FlowAccumulator(grid)
        >>> fs = FastscapeEroder(grid, K_sp=0.01)

        Run this model for 300 100-year timesteps.

        >>> dt = 100.0
        >>> for i in range(300):
        ...     nf.run_one_step(dt)
        ...     fr.run_one_step()
        ...     fs.run_one_step(dt)
        >>> z.reshape(grid.shape).round(decimals=2)
        array([[  0.  ,   0.  ,   0.  ,   0.  ,   0.  ,   0.  ],
               [  5.  ,   5.  ,   0.  ,   0.  ,   0.  ,   0.  ],
               [  7.39,   7.38,   2.38,   2.89,   0.  ,   0.  ],
               [  9.36,  11.43,   5.51,   6.42,   3.54,   3.54],
               [ 15.06,  15.75,  10.6 ,  11.42,   8.54,   8.54],
               [ 15.06,  15.06,  10.7 ,  11.42,   8.54,   8.54]])

        The faulted nodes have been uplifted and eroded! Note that here the
        boundary nodes are also uplifted.

        If a user chooses to set a rate-time pattern, but not run NormalFault
        along with all other components, then NormalFault's internal
        time-keeping will be incorrect.

        For example, if a user wanted to only run NormalFault every tenth
        timestep (or some more seismogenically reasonable set of times).

        >>> grid = RasterModelGrid((6, 6), spacing=10)
        >>> z = grid.add_zeros('node', 'topographic__elevation')
        >>> nf = NormalFault(grid, **param_dict)
        >>> fr = FlowAccumulator(grid)
        >>> fs = FastscapeEroder(grid, K_sp=0.01)
        >>> model_time = 0.0
        >>> dt = 100.0
        >>> for i in range(300):
        ...     if i%10 == 0:
        ...         nf.run_one_step(dt, current_time = model_time)
        ...     fr.run_one_step()
        ...     fs.run_one_step(dt)
        ...     model_time += dt
        >>> model_time
        30000.0
        >>> nf.current_time
        29100.0

        This may seem incorrect as the model time and the current time are not
        the same. However, this is correct as the NormalFault runs one step only
        at the first of each 10 timesteps. The value of 2910.0 reflects that
        NormalFault ran from model time 29000.0 to 29100.0 and has not run since
        which is what we'd expect.

        """
        # call the class Normal Fault inherits from
        super(NormalFault, self).__init__(grid)

        # save a reference to the grid
        self._grid = grid

        # get the surface to be faulted
        self.surfaces = {}
        self._not_yet_instantiated = []
        if isinstance(faulted_surface, list):
            # if faulted surface is a list, then itterate through multiple
            # surfaces and save
            for surf in faulted_surface:
                try:
                    self.surfaces[surf] = grid.at_node[surf]
                except FieldError:
                    self._not_yet_instantiated.append(surf)
        else:
            self.surfaces[faulted_surface] = grid.at_node[faulted_surface]

        if fault_dip_angle > 90.0:
            raise ValueError('NormaFault fault_dip_angle must be less than 90 '
                             'degrees.')

        # get the fault throw parameter values from the parameter dictionary
        self.throw_time = np.array(fault_throw_rate_through_time['time'])
        self.throw_rate = np.array(fault_throw_rate_through_time['rate'])
        self.fault_dip = np.deg2rad(fault_dip_angle)
        self.uplift = self.throw_rate * np.sin(self.fault_dip)

        # Identify in current boundaries will be included
        self.include_boundaries = include_boundaries

        # Instantiate record of current time.
        self.current_time = 0.0

        # get the fault trace dictionary and use to to calculate where the
        # faulted nodes are located.
        self.fault_trace = fault_trace
        dx = self.fault_trace['x2'] - self.fault_trace['x1']
        dy = self.fault_trace['y2'] - self.fault_trace['y1']
        self.fault_azimuth = np.mod(np.arctan2(dy, dx), TWO_PI)
        self.fault_anti_azimuth = self.fault_azimuth + np.pi
        # deal with the edge case in which dx == 0
        if dx == 0:
            self.dy_over_dx = 0.0
            self.fault_trace_y_intercept = 0.0
            self.fault_trace_x_intercept = self.fault_trace['x2']
        else:
            self.dy_over_dx = dy/dx
            self.fault_trace_y_intercept = self.fault_trace['y1'] - (self.dy_over_dx * self.fault_trace['x1'])
            self.fault_trace_x_intercept = 0.0

        # set the considered nodes based on whether the boundaries will be
        # included in the faulted terrain.
        if self.include_boundaries:
            potential_nodes = np.arange(self._grid.size('node'))
        else:
            potential_nodes = self._grid.core_nodes

        # select those nodes that are on the correct side of the fault
        dx_pn = (self._grid.x_of_node[potential_nodes] - self.fault_trace_x_intercept)
        dy_pn = (self._grid.y_of_node[potential_nodes] - self.fault_trace_y_intercept)
        potential_angles = np.mod(np.arctan2(dy_pn, dx_pn), TWO_PI)
        if self.fault_anti_azimuth <= TWO_PI:
            faulted_node_ids = potential_nodes[((potential_angles>self.fault_azimuth) &
                                                  (potential_angles <= (self.fault_anti_azimuth)))]
        else:
            faulted_node_ids = potential_nodes[((potential_angles>self.fault_azimuth) |
                                                  (potential_angles <= np.mod(self.fault_anti_azimuth, TWO_PI)))]

        # save a n-node array of boolean identifing faulted nodes.
        self.faulted_nodes = np.zeros(self._grid.size('node'), dtype=bool)
        self.faulted_nodes[faulted_node_ids] = True

    def _check_surfaces(self):
        if len(self._not_yet_instantiated)>0:
            still_not_instantiated = []
            for surf in self._not_yet_instantiated:
                if surf in self._grid.at_node:
                    self.surfaces[surf] = self._grid.at_node[surf]
                else:
                    still_not_instantiated.append(surf)
            self._not_yet_instantiated = still_not_instantiated

    def run_one_earthquake(self, dz):
        """ """
        self._check_surfaces()

        # save z before uplift only if using include boundaries.
        if self.include_boundaries:
            surfs_before_uplift = {}
            for surf_name in self.surfaces:
                surfs_before_uplift[surf_name] = self.surfaces[surf_name].copy()

        # uplift the faulted_nodes
        for surf_name in self.surfaces:
            self.surfaces[surf_name][self.faulted_nodes] += dz

        # if faulted nodes includes boundaries we must do some extra work because
        # landlab components will typically not erode these boundaries. This means
        # they will be uplifted but not eroded.

        if self.include_boundaries:

            #  here our goal is to set faulted boundaries to average of open
            # node faulted neighbors

            # create boolean of the faulted boundary nodes
            faulted_boundaries = self.faulted_nodes.copy()
            faulted_boundaries[self._grid.core_nodes] = False

            core_nodes = np.zeros(self._grid.size('node'), dtype=bool)
            core_nodes[self._grid.core_nodes] = True

            neighbor_is_core = core_nodes[self._grid.neighbors_at_node]
            neighbor_is_faulted = self.faulted_nodes[self._grid.neighbors_at_node]

            neighbor_for_averaging = neighbor_is_faulted&neighbor_is_core

            # Identify nodes that have at least one adjacent node that is both
            # faulted and not a boundary node.
            # average the pre-uplift topography on those adjacent nodes and assign
            # to the boundary node.
            # here we use the pre-uplift elevation because other steps in the model
            # may diminish this topography.

            averaged = neighbor_for_averaging[faulted_boundaries].sum(axis=1) == 1
            if any(averaged):
                averaged_nodes = np.where(faulted_boundaries)[0][np.where(averaged)[0]]
                for surf_name in self.surfaces:
                    elevations_to_average =  surfs_before_uplift[surf_name][self._grid.neighbors_at_node]
                    elevations_to_average[self._grid.neighbors_at_node == -1] = np.nan
                    elevations_to_average[neighbor_for_averaging == False] = np.nan
                    self.surfaces[surf_name][averaged_nodes] = np.nanmean(elevations_to_average[averaged_nodes], axis=1)

            # identify any boundary nodes that are not being averaged. This will
            # happen at the corners on RasterModelGrids. Average over adjacent
            # nodes that are faulted. These nodes will be boundary nodes.
            # here we use the current topography as we will have just updated the
            # adjacent nodes in the prior block.
            if any(averaged == False):
                un_averaged_nodes = np.where(faulted_boundaries)[0][np.where(averaged == False)[0]]
                for surf_name in self.surfaces:
                    elevations_to_average =  self.surfaces[surf_name][self._grid.neighbors_at_node]
                    elevations_to_average[self._grid.neighbors_at_node == -1] = np.nan
                    elevations_to_average[neighbor_is_faulted == False] = np.nan
                    self.surfaces[surf_name][un_averaged_nodes] = np.nanmean(elevations_to_average[un_averaged_nodes], axis=1)

    def run_one_step(self, dt, current_time=None):
        """Run_one_step method for NormalFault.

        Parameters
        ----------
        dt : float
            Time increment used to advance the NormalFault component.
        current_time : float, optional
            If NormalFault is not being advanced by dt every timestep with all
            components, its internal time may be incorrect, this can be remedied
            by providing a value for current time. Default value is None which
            results in the internal timekeeping not being changed.
        """
        # if current time is provided, use it to re-set the current time.
        if current_time is not None:
            self.current_time = current_time

        # calculate the current uplift rate
        current_uplift_rate = np.interp(self.current_time, self.throw_time, self.throw_rate)

        # run one earthquake of size current_uplift_rate * dt
        self.run_one_earthquake(current_uplift_rate * dt)

        # increment time
        self.current_time += dt



"""
These tests use the following arrangement of nodes::

          o --- o --- o --- o
         /     /     /     /
       o --- o --- o --- o
      /     /     /    /
    o --- o --- o --- o

This result in two bound voronoi polygons (12 total) centered on points
5 and 6.
"""

from numpy.testing import assert_array_equal, assert_array_almost_equal
from scipy.spatial import Voronoi
import numpy as np

from landlab.graph.voronoi.voronoi_helpers import VoronoiConverter

#     get_finite_regions, is_patch, is_link, get_patch_at_region,
#     get_link_at_ridge, get_node_at_vertex, get_patches_at_link,
#     get_nodes_at_link, get_nodes, get_ridges_at_region,
#     get_links_at_patch, get_corner_at_patch)


NODE_X = (0. , 1. , 2. , 3. ,
           .1, 1.1, 2.1, 3.1,
           .2, 1.2, 2.2, 3.2)
NODE_Y = (0. , 0. , 0. , 0. ,
          1. , 1. , 1. , 1. ,
          2. , 2. , 2. , 2.)
POINTS = list(zip(NODE_X, NODE_Y))


def test_get_finite_regions():
    """Test the regions that are bound."""
    converter = VoronoiConverter(Voronoi(POINTS))

    regions = converter.get_finite_regions()

    assert len(regions) == 13 # There's one extra "empty" region
    assert len(regions[regions == 0]) == 11
    assert len(regions[regions == 1]) == 2


def test_is_patch():
    """Test if a region is a patch."""
    converter = VoronoiConverter(None)

    assert converter.is_patch([1, 2, 3])
    assert not converter.is_patch([1, 2, 3, -1])
    assert not converter.is_patch([])


def test_is_link():
    """Test if a ridge is a link."""
    v = Voronoi(POINTS)
    converter = VoronoiConverter(v)

    links = [converter.is_link(r) for r in range(len(v.ridge_vertices))]

    assert sum(links) == 11


def test_patch_at_region():
    """Test mapping voronoi regions to patches."""
    v = Voronoi(POINTS)
    converter = VoronoiConverter(v)

    patch_at_region = converter.get_patch_at_region()

    assert patch_at_region.shape == (len(v.regions), )
    assert sum(patch_at_region >= 0) == 2


def test_link_at_ridge():
    """Test mapping voronoi ridges to links."""
    v = Voronoi(POINTS)
    converter = VoronoiConverter(v)

    link_at_ridge = converter.get_link_at_ridge()

    assert link_at_ridge.shape == (len(v.ridge_vertices), )
    assert sum(link_at_ridge >= 0) == 11


def test_patches_at_link():
    """Test getting link patches."""
    converter = VoronoiConverter(Voronoi(POINTS))

    patches_at_link = converter.get_patches_at_link()

    assert patches_at_link.shape == (11, 2)
    for patches in patches_at_link:
        assert patches[0] != -1 or patches[1] != -1


def test_node_at_vertex():
    """Test mapping voronoi vertices to nodes."""
    v = Voronoi(POINTS)
    converter = VoronoiConverter(v)

    node_at_vertex = converter.get_node_at_vertex()

    assert node_at_vertex.shape == (len(v.vertices), )
    assert sum(node_at_vertex >= 0) == 10

    node_at_vertex = node_at_vertex[node_at_vertex >= 0]
    node_at_vertex.sort()
    assert_array_equal(node_at_vertex, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])


def test_nodes_at_link():
    """Test getting nodes at links from a voronoi."""
    converter = VoronoiConverter(Voronoi(POINTS))

    nodes_at_link = converter.get_nodes_at_link()

    assert nodes_at_link.shape == (11, 2)
    assert np.all(nodes_at_link >= 0)
    assert np.all(nodes_at_link < 10)

    assert_array_equal(np.unique(nodes_at_link),
                       [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])


def test_nodes():
    """Test getting nodes at links from a voronoi."""
    converter = VoronoiConverter(Voronoi(POINTS))

    nodes = converter.get_nodes()

    assert nodes.shape == (10, 2)
    assert isinstance(nodes[0, 0], float)


def test_ridges_at_region():
    """Test getting ridges that bound regions."""
    v = Voronoi(POINTS)
    converter = VoronoiConverter(v)

    ridges_at_region = converter.get_ridges_at_region()

    assert ridges_at_region.shape == (len(v.regions), 6)
    assert isinstance(ridges_at_region[0, 0], np.int_)
    assert_array_equal(ridges_at_region[0], [-1] * 6)


def test_links_at_patch():
    """Test getting links that bound patches from a voronoi."""
    converter = VoronoiConverter(Voronoi(POINTS))

    links_at_patch = converter.get_links_at_patch()

    assert links_at_patch.shape == (2, 6)
    assert np.all(links_at_patch >= 0)
    assert np.all(links_at_patch < 11)


def test_corner_at_patch():
    """Test getting corners for each patch."""
    v = Voronoi(POINTS)
    converter = VoronoiConverter(v)

    corner_at_patch = converter.get_corner_at_patch()

    assert corner_at_patch.shape == (2, )
    assert np.all(corner_at_patch >= 0)
    assert np.all(corner_at_patch < len(v.points))



#! /usr/bin/env python
"""
Unit tests for landlab.model_parameter_dictionary
"""
import pytest
import os
import tempfile
import numpy as np

from landlab import ModelParameterDictionary
from landlab.core.model_parameter_dictionary import MissingKeyError, ParameterValueError


def test_read_file(pdict_setup):
    all_keys = set(
        ["FLOAT_VAL", "INT_VAL", "STRING_VAL", "TRUE_BOOL_VAL", "FALSE_BOOL_VAL"]
    )
    param_list = set(pdict_setup.param_dict.params())

    assert param_list == all_keys


def test_read_file_name(pdict_setup):
    (prm_fd, prm_file_name) = tempfile.mkstemp()

    prm_file = os.fdopen(prm_fd, "w")
    prm_file.write(pdict_setup.param_dict_file)
    prm_file.close()

    param_dict = ModelParameterDictionary()
    param_dict.read_from_file(prm_file_name)

    os.remove(prm_file_name)

    all_keys = set(
        ["FLOAT_VAL", "INT_VAL", "STRING_VAL", "TRUE_BOOL_VAL", "FALSE_BOOL_VAL"]
    )
    param_list = set(param_dict.params())

    assert param_list == all_keys


def test_read_file_like_twice(pdict_setup):
    from six import StringIO

    param_file = StringIO(pdict_setup.param_dict_file)
    param_dict_1 = ModelParameterDictionary()
    param_dict_2 = ModelParameterDictionary()

    param_dict_1.read_from_file(param_file)
    param_dict_2.read_from_file(param_file)


def test_read_int(pdict_setup):
    assert pdict_setup.param_dict.read_int("INT_VAL") == 1

    with pytest.raises(ParameterValueError):
        pdict_setup.param_dict.read_int("FLOAT_VAL")

    with pytest.raises(MissingKeyError):
        pdict_setup.param_dict.read_int("MISSING_INT")

    assert pdict_setup.param_dict.read_int("MISSING_INT", 2) == 2


def test_get_int(pdict_setup):
    assert pdict_setup.param_dict.get("INT_VAL", ptype=int) == 1

    with pytest.raises(ParameterValueError):
        pdict_setup.param_dict.get("FLOAT_VAL", ptype=int)

    with pytest.raises(MissingKeyError):
        pdict_setup.param_dict.get("MISSING_INT", ptype=int)


def test_set_default(pdict_setup):
    pdict_setup.param_dict.setdefault("MISSING_INT", 2)
    assert pdict_setup.param_dict.read_int("MISSING_INT") == 2


def test_read_float(pdict_setup):
    assert pdict_setup.param_dict.read_float("FLOAT_VAL") == 2.2
    assert pdict_setup.param_dict.read_float("INT_VAL") == 1

    with pytest.raises(ParameterValueError):
        pdict_setup.param_dict.read_float("STRING_VAL")

    with pytest.raises(MissingKeyError):
        pdict_setup.param_dict.read_float("MISSING_FLOAT")


def test_read_string(pdict_setup):
    assert pdict_setup.param_dict.read_string("STRING_VAL") == "The Landlab"
    assert pdict_setup.param_dict.read_string("INT_VAL") == "1"
    assert pdict_setup.param_dict.read_string("FLOAT_VAL") == "2.2"

    with pytest.raises(MissingKeyError):
        pdict_setup.param_dict.read_string("MISSING_STRING")


def test_read_bool(pdict_setup):
    assert pdict_setup.param_dict.read_bool("TRUE_BOOL_VAL") == True
    assert pdict_setup.param_dict.read_bool("FALSE_BOOL_VAL") == False

    with pytest.raises(MissingKeyError):
        pdict_setup.param_dict.read_bool("MISSING_BOOLEAN")

    with pytest.raises(ParameterValueError):
        pdict_setup.param_dict.read_bool("STRING_VAL")


def test_dict_keys(pdict_setup):
    all_keys = set(
        ["FLOAT_VAL", "INT_VAL", "STRING_VAL", "TRUE_BOOL_VAL", "FALSE_BOOL_VAL"]
    )
    assert set(pdict_setup.param_dict) == all_keys
    for key in all_keys:
        assert key in pdict_setup.param_dict


def test_dict_index(pdict_setup):
    assert pdict_setup.param_dict["INT_VAL"] == "1"


def test_auto_type(auto_type_setup):
    assert auto_type_setup["INT_VAL"] == 1
    assert auto_type_setup["DBL_VAL"] == 1.2
    assert auto_type_setup["STR_VAL"] == "landlab"
    assert auto_type_setup["BOOL_VAL"] == True


def test_int_vector(auto_type_setup):
    val = auto_type_setup["INT_ARRAY_VAL"]

    assert list(val) == [1, 2, 4, 7]
    assert isinstance(val, np.ndarray)
    assert val.dtype == np.int


def test_float_vector(auto_type_setup):
    val = auto_type_setup["DBL_ARRAY_VAL"]

    assert list(val) == [1., 2., 4., 7.]
    assert isinstance(val, np.ndarray)
    assert val.dtype == np.float



#! /usr/bin/env python
"""
Utility functions for loading components for The Landlab.
"""

import os
from landlab.framework.interfaces import BmiBase


_COMPONENT_PATH = [
    os.path.join(os.path.dirname(__file__), '..', 'components'),
]

try:
    paths = os.environ['LANDLAB_PATH'].split(os.pathsep)
except KeyError:
    pass
else:
    _COMPONENT_PATH = paths + _COMPONENT_PATH


def iscomponent(value, cls):
    """
    Check if *value* is a component for The Landlab. *value* is a component
    if it implements the *cls* or it is an instance of *cls*.

    Returns ``True`` if *value* is a component, otherwise ``False``.
    """
    try:
        return (cls in value.__implements__ or
                cls.__name__ in value.__implements__ or
                isinstance(value, cls))
    except AttributeError:
        return False


def load_components_from_dir(path, cls):
    """
    Look for components for Landlab in *path*. Identify components as being an
    instance of *cls*. Returns a dictionary of discovered component names as
    keys and component classes as values.
    """
    import sys
    import imp

    components = {}

    sys.path.insert(0, path)
    cwd = os.getcwd()

    os.chdir(path)
    for file_name in os.listdir('.'):
        if os.path.isfile(file_name) and file_name.endswith('.py'):
            (mod_name, _) = os.path.splitext(file_name)
            mod = imp.load_module(mod_name, *imp.find_module(mod_name))
            for (name, value) in mod.__dict__.items():
                if iscomponent(value, cls):
                    components[name] = value

    os.chdir(cwd)
    sys.path.pop(0)

    return components


def load_components(cls, paths=None):
    """
    Load components from a series of directories.

    Components found earlier in the search path order override those
    discovered later. Use the *paths* keyword to specify a list of paths to
    search for components.

    .. seealso::

        :func:`load_components_from_dir`
    """
    if not paths:
        paths = _COMPONENT_PATH

    components = {}
    for path in paths[::-1]:
        components.update(load_components_from_dir(path, cls))
    return components


def load_landlab_components(paths=None):
    """
    Load components for The Landlab. These are classes that implement BmiBase.
    See :func:`load_components_from_dir` for the meaning of *paths* keyword.

    .. seealso::

        :func:`load_components_from_dir`
    """
    return load_components(BmiBase, paths=paths)



"""Plot drainage network.

"""
# KRB, FEB 2017.
import six
from landlab import CORE_NODE, FIXED_VALUE_BOUNDARY, FIXED_GRADIENT_BOUNDARY, CLOSED_BOUNDARY
import matplotlib.pylab as plt
from landlab.plot.imshow import imshow_node_grid
import numpy as np

def drainage_plot(mg, 
                  surface='topographic__elevation', 
                  receivers=None, 
                  proportions=None, 
                  surf_cmap='gray',
                  quiver_cmap='viridis',
                  title = 'Drainage Plot'):
    

    if isinstance(surface, six.string_types):
        colorbar_label = surface
    else:
        colorbar_label = 'topographic_elevation'
    imshow_node_grid(mg, surface, cmap=surf_cmap, colorbar_label=colorbar_label)
    
    if receivers is None:
        try:
            receivers = mg.at_node['flow__receiver_nodes']
            if proportions is None:
                try:
                    proportions = mg.at_node['flow__receiver_proportions']
                except:
                    pass
        except: 
            receivers = np.reshape(mg.at_node['flow__receiver_node'],(mg.number_of_nodes,1))
        
    nreceievers = int(receivers.size/receivers.shape[0])
    
    propColor=plt.get_cmap(quiver_cmap)

    for j in range(nreceievers):
        rec = receivers[:,j]
        is_bad = rec == -1
        
        xdist =  -0.8*(mg.node_x-mg.node_x[rec])
        ydist =  -0.8*(mg.node_y-mg.node_y[rec])
        
        if proportions is None:
           proportions = np.ones_like(receivers, dtype=float)
        
        is_bad[proportions[:,j]==0.]=True
        
        xdist[is_bad] = np.nan
        ydist[is_bad] = np.nan
             
        prop = proportions[:,j]*256.
        lu = np.floor(prop)
        colors = propColor(lu.astype(int))
        
        shape = (mg.number_of_nodes, 1)
        
        plt.quiver(mg.node_x.reshape(shape), mg.node_y.reshape(shape), 
               xdist.reshape(shape), ydist.reshape(shape), 
               color=colors,
               angles='xy',
               scale_units='xy', 
               scale=1,
               zorder=3) 
    
    # Plot differen types of nodes:
    o, = plt.plot(mg.node_x[mg.status_at_node == CORE_NODE], mg.node_y[mg.status_at_node == CORE_NODE], 'b.', label='Core Nodes', zorder=4)
    fg, = plt.plot(mg.node_x[mg.status_at_node == FIXED_VALUE_BOUNDARY], mg.node_y[mg.status_at_node == FIXED_VALUE_BOUNDARY], 'c.', label='Fixed Gradient Nodes', zorder=5)
    fv, = plt.plot(mg.node_x[mg.status_at_node == FIXED_GRADIENT_BOUNDARY], mg.node_y[mg.status_at_node ==FIXED_GRADIENT_BOUNDARY], 'g.', label='Fixed Value Nodes', zorder=6)
    c, = plt.plot(mg.node_x[mg.status_at_node == CLOSED_BOUNDARY], mg.node_y[mg.status_at_node ==CLOSED_BOUNDARY], 'r.', label='Closed Nodes', zorder=7)

    node_id = np.arange(mg.number_of_nodes)
    flow_to_self = receivers[:,0]==node_id
                            
    fts, = plt.plot(mg.node_x[flow_to_self], mg.node_y[flow_to_self], 'kx', markersize=6, label = 'Flows To Self', zorder=8)
    
    ax = plt.gca()
    
    ax.legend(labels = ['Core Nodes', 'Fixed Gradient Nodes', 'Fixed Value Nodes', 'Closed Nodes', 'Flows To Self'],
              handles = [o, fg, fv, c, fts], numpoints=1, loc='center left', bbox_to_anchor=(1.7, 0.5))
    sm = plt.cm.ScalarMappable(cmap=propColor, norm=plt.Normalize(vmin=0, vmax=1))
    sm._A = []
    cx = plt.colorbar(sm)
    cx.set_label('Proportion of Flow')
    plt.title(title)
    plt.show()


import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal

from landlab import RasterModelGrid


def test_unit_spacing():
    grid = RasterModelGrid((4, 5))
    assert_array_equal(grid.length_of_link, np.ones(31))


def test_non_unit_spacing():
    grid = RasterModelGrid((4, 5), spacing=(3, 4))
    assert_array_equal(grid.length_of_link,
                       [4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.])


def test_link_length():
    grid = RasterModelGrid((4, 5), spacing=(3, 4))
    assert_array_equal(grid.length_of_link,
                       [4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.])


def test_active_link_length():
    grid = RasterModelGrid((4, 5), spacing=(3, 4))
    assert_array_equal(grid.length_of_link[grid.active_links],
                       [3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3.])



#!/usr/bin/env python

import numpy as np
import six

from landlab import Component, CLOSED_BOUNDARY
from ...utils.decorators import use_file_name_or_kwds


class SoilInfiltrationGreenAmpt(Component):

    """Infiltrate surface water into a soil following the Green-Ampt method.

    This code is based on an overland flow model by Francis Rengers and
    colleagues, after Julien et al., 1995. The infiltration scheme follows the
    Green and Ampt equation.

    It was implemented in Landlab by DEJH, March '16. Please cite
    Rengers et al., 2016, Model Predictions of Water Runoff in Steep
    Catchments after Wildfire, WRR.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import SoilInfiltrationGreenAmpt
    >>> mg = RasterModelGrid((4,3), spacing=10.)
    >>> hydraulic_conductivity = mg.ones('node')*1.e-6
    >>> hydraulic_conductivity.reshape((4,3))[0:2,:] *= 10000.
    >>> h = mg.add_ones('node', 'surface_water__depth')
    >>> h *= 0.01
    >>> d = mg.add_ones('node', 'soil_water_infiltration__depth', dtype=float)
    >>> d *= 0.2
    >>> SI = SoilInfiltrationGreenAmpt(
    ...     mg,hydraulic_conductivity=hydraulic_conductivity)
    >>> for i in range(10):  # 100s total
    ...     SI.run_one_step(10.)
    >>> mg.at_node['surface_water__depth']
    array([  1.00000000e-08,   1.00000000e-08,   1.00000000e-08,
             1.00000000e-08,   1.00000000e-08,   1.00000000e-08,
             9.88530416e-03,   9.88530416e-03,   9.88530416e-03,
             9.88530416e-03,   9.88530416e-03,   9.88530416e-03])
    >>> mg.at_node['soil_water_infiltration__depth']
    array([ 0.20999999,  0.20999999,  0.20999999,  0.20999999,  0.20999999,
            0.20999999,  0.2001147 ,  0.2001147 ,  0.2001147 ,  0.2001147 ,
            0.2001147 ,  0.2001147 ])
    """

    _name = 'SoilInfiltrationGreenAmpt'

    _input_var_names = (
        'surface_water__depth',
        'soil_water_infiltration__depth'
        )

    _output_var_names = (
        'surface_water__depth',
        'soil_water_infiltration__depth',
        )

    _var_units = {
        'surface_water__depth': 'm',
        'soil_water_infiltration__depth': 'm'
        }

    _var_mapping = {
        'surface_water__depth': 'node',
        'soil_water_infiltration__depth': 'node'
        }

    _var_doc = {
        'surface_water__depth': 'Depth of water above the surface',
        'soil_water_infiltration__depth': (
            'Water column height above the surface previously absorbed into '
            'the soil. Note that this is NOT the actual depth of the wetted '
            'front, which also depends on the porosity.')
        }

    # This follows mean values from Rawls et al., 1992; lambda then h_b
    SOIL_PROPS = {
        'sand': (0.694, 0.0726),
        'loamy sand': (0.553, 0.0869),
        'sandy loam': (0.378, 0.1466),
        'loam': (0.252, 0.1115),
        'silt loam': (0.234, 0.2076),
        'sandy clay loam': (0.319, 0.2808),
        'clay loam': (0.242, 0.2589),
        'silty clay loam': (0.177, 0.3256),
        'sandy clay': (0.223, 0.2917),
        'silty clay': (0.150, 0.3419),
        'clay': (0.165, 0.3730),
    }

    @use_file_name_or_kwds
    def __init__(self, grid, hydraulic_conductivity=0.005,
                 soil_bulk_density=1590., rock_density=2650.,
                 initial_soil_moisture_content=0.15, soil_type='sandy loam',
                 volume_fraction_coarse_fragments=0.2,
                 coarse_sed_flag=False,
                 surface_water_minimum_depth=1.e-8,
                 soil_pore_size_distribution_index=None,
                 soil_bubbling_pressure=None,
                 wetting_front_capillary_pressure_head=None, **kwds):
        """Kinematic wave approximation overland flow component.

        Parameters
        ----------
        grid : RasterModelGrid
            A grid.
        hydraulic_conductivity : float, array, or field name (m/s)
            The soil effective hydraulic conductivity.
        soil_bulk_density : float (kg/m**3)
            The dry bulk density of the soil.
        rock_density : float (kg/m**3)
            The density of the soil constituent material (i.e., lacking porosity).
        initial_soil_moisture_content : float (m**3/m**3, 0. to 1.)
            The fraction of the initial pore space filled with water.
        soil_type : {'sand', loamy sand', 'sandy loam', 'loam', 'silt loam',
                     'sandy clay loam', 'clay loam', 'silty clay loam',
                     'sandy clay', 'silty clay', 'clay'}, or None
            A soil type to automatically set soil_pore_size_distribution_index
            and soil_bubbling_pressure, using mean values from Rawls et al.,
            1992.
        volume_fraction_coarse_fragments : float (m**3/m**3, 0. to 1.)
            The fraction of the soil made up of rocky fragments with very
            little porosity, with diameter > 2 mm.
        coarse_sed_flag : boolean, optional
            If this flag is set to true, the fraction of coarse material in the
            soil column with be used as a correction for phi, the porosity factor.
        surface_water_minimum_depth : float (m), optional
            A minimum water depth to stabilize the solutions for surface flood
            modelling. Leave as the default in most normal use cases.
        soil_pore_size_distribution_index : float, optional
            An index describing the distribution of pore sizes in the soil,
            and controlling effective hydraulic conductivity at varying water
            contents, following Brooks and Corey (1964). Can be set by
            soil_type. Typically denoted "lambda".
        soil_bubbling_pressure : float (m), optional
            The bubbling capillary pressure of the soil, controlling effective
            hydraulic conductivity at varying water contents, following Brooks
            and Corey (1964). Can be set by soil_type. Typically denoted "h_b".
        wetting_front_capillary_pressure_head : float (m), optional
            The effective head at the wetting front in the soil driven by
            capillary pressure in the soil pores. If not set, will be
            calculated by the component from the pore size distribution and
            bubbling pressure, following Brooks and Corey.
        """

        self._grid = grid
        self.min_water = surface_water_minimum_depth
        self.hydraulic_conductivity = hydraulic_conductivity

        if not coarse_sed_flag:
            volume_fraction_coarse_fragments = 0.

        self.moisture_deficit = self.calc_moisture_deficit(
            soil_bulk_density=soil_bulk_density,
            rock_density=rock_density,
            volume_fraction_coarse_fragments=volume_fraction_coarse_fragments,
            soil_moisture_content=initial_soil_moisture_content)

        if wetting_front_capillary_pressure_head is None:
            self.capillary_pressure = self.calc_soil_pressure(
                soil_type=soil_type,
                soil_pore_size_distribution_index=soil_pore_size_distribution_index,
                soil_bubbling_pressure=soil_bubbling_pressure)
        else:
            self.capillary_pressure = wetting_front_capillary_pressure_head

    @staticmethod
    def calc_soil_pressure(soil_type=None,
                           soil_pore_size_distribution_index=1.,
                           soil_bubbling_pressure=0.):
        """Calculate capillary pressure in a soil type.

        Parameters
        ----------
        soil_type : str, optional
            The name of a soil type.
        soil_pore_size_distribution_index : float
            Pore-size distribution index [-].
        soil_bubbling_pressure : float
            Bubbling pressure [m].
        """
        if soil_type is None:
            soil_props = (soil_pore_size_distribution_index,
                          soil_bubbling_pressure)
        else:
            try:
                soil_props = SoilInfiltrationGreenAmpt.SOIL_PROPS[soil_type]
            except KeyError:
                raise ValueError('{0}: unknown soil type'.format(soil_type))

        return SoilInfiltrationGreenAmpt.calc_pressure_head(*soil_props)

    @staticmethod
    def calc_pressure_head(lam, h_b):
        """Calculate pressure head.

        Pressure head is set using *lambda* and *h_b*, using an
        equation after Brooks-Corey (1964), following Rawls et al., 1992.

        Parameters
        ----------
        lam : float, optional
            Pore-size distribution index. Exponent that describes the
            distribution of pore sizes in the soil, and controls
            effective hydraulic conductivity at varying water
            contents, following Brooks and Corey (1964) [-].
        h_b : float (m), optional
            Bubbling pressure. Capillary pressure of the soil,
            controlling effective hydraulic conductivity at varying
            water contents, following Brooks and Corey (1964) [m]
        """
        return ((2. + 3. * lam) / (1. + 3. * lam) * h_b * .5)

    @staticmethod
    def calc_moisture_deficit(soil_bulk_density=1590., rock_density=2650.,
                              volume_fraction_coarse_fragments=0.,
                              soil_moisture_content=0.):
        """Calculate the moisture deficit in a soil.
        
        Parameters
        ----------
        soil_bulk_density : float or array of float
            Bulk density of the soil [kg / m3].
        rock_density : float or array of float
            Density of rock [kg / m3].
        volume_fraction_coarse_fragments : float or array of float
            Volume fraction of sediment made up of coarse grains [-].
        soil_moisture_content : float or array of float
            Fraction of soil filled with water [-].

        Returns
        -------
        float or array of float
            Moisture deficit.
        """
        if np.any(soil_bulk_density <= 0.):
            raise ValueError('non-positive soil bulk density')
        if np.any(rock_density < soil_bulk_density):
            raise ValueError('soil bulk density greater than rock density')
        if np.any(volume_fraction_coarse_fragments < 0.):
            raise ValueError('negative volume fraction of coarse grains')
        if np.any(volume_fraction_coarse_fragments > 1.):
            raise ValueError('volume fraction of coarse grains')

        soil_porosity = 1. - np.true_divide(soil_bulk_density, rock_density)
        soil_porosity *= 1. - volume_fraction_coarse_fragments

        if np.any(soil_moisture_content > soil_porosity):
            raise ValueError('soil moisture greater than porosity')

        return soil_porosity - soil_moisture_content

    @property
    def min_water(self):
        """Minimum surface water depth."""
        return self._min_water

    @min_water.setter
    def min_water(self, new_value):
        if np.any(new_value <= 0.):
            raise ValueError('minimum water depth must be positive')
        self._min_water = new_value

    @property
    def hydraulic_conductivity(self):
        """Hydraulic conductivity of soil."""
        return self._hydraulic_conductivity
    
    @hydraulic_conductivity.setter
    def hydraulic_conductivity(self, new_value):
        if isinstance(new_value, six.string_types):
            new_value = self.grid.at_node[new_value]
        if np.any(new_value < 0.):
            raise ValueError('hydraulic conductivity must be positive')
        self._hydraulic_conductivity = new_value

    @property
    def moisture_deficit(self):
        """Moisture deficit of soil."""
        return self._moisture_deficit

    @moisture_deficit.setter
    def moisture_deficit(self, new_value):
        if np.any(new_value < 0.):
            raise ValueError('negative moisture deficit')
        self._moisture_deficit = new_value

    @property
    def capillary_pressure(self):
        """Capillary pressure of soil."""
        return self._capillary_pressure

    @capillary_pressure.setter
    def capillary_pressure(self, new_value):
        if np.any(new_value < 0.):
            raise ValueError('negative capillary pressure')
        self._capillary_pressure = new_value

    def run_one_step(self, dt):
        """Update fields with current hydrologic conditions.

        Parameters
        ----------
        dt : float (s)
            The imposed timestep for the model.
        """
        water_depth = self.grid.at_node['surface_water__depth']
        infiltration_depth = self.grid.at_node['soil_water_infiltration__depth']

        assert(np.all(infiltration_depth >= 0.))

        wettingfront_depth = infiltration_depth / self.moisture_deficit
        
        potential_infilt = dt * self.hydraulic_conductivity * (
            (wettingfront_depth + self.capillary_pressure + water_depth) /
            wettingfront_depth)
        np.clip(potential_infilt, 0., None, out=potential_infilt)

        available_water = water_depth - self.min_water
        np.clip(available_water, 0., None, out=available_water)

        actual_infiltration = np.choose(potential_infilt > available_water,
                                        (potential_infilt, available_water))

        water_depth -= actual_infiltration
        infiltration_depth += actual_infiltration



import pytest
import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_default_names():
    rmg = RasterModelGrid((4, 5))
    assert rmg.axis_name == ('y', 'x')


def test_name_keyword():
    rmg = RasterModelGrid(
        (4, 5),
        axis_name=['longitude', 'latitude'],
        axis_units=['degrees_east', 'degrees_north'],
    )
    assert rmg.axis_name == ('longitude', 'latitude')


def test_name_setter():
    rmg = RasterModelGrid(
        (4, 5),
        axis_name=['longitude', 'latitude'],
        axis_units=['degrees_east', 'degrees_north'],
    )
    rmg.axis_name = ('yyy', 'xxx')
    assert rmg.axis_name == ('yyy', 'xxx')


def test_name_setter_too_few_names():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.axis_name = ('z', )


def test_name_setter_too_many_names():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.axis_name = ('z', 'y', 'x')


def test_default_units():
    rmg = RasterModelGrid((4, 5))
    assert rmg.axis_units == ('-', '-')


def test_name_keyword():
    rmg = RasterModelGrid(
        (4, 5),
        axis_name=['longitude', 'latitude'],
        axis_units=['degrees_east', 'degrees_north'],
    )
    assert rmg.axis_units == ('degrees_east', 'degrees_north')


def test_name_setter():
    rmg = RasterModelGrid(
        (4, 5),
        axis_name=['longitude', 'latitude'],
        axis_units=['degrees_east', 'degrees_north'],
    )
    rmg.axis_units = ('mm', 'cm')
    assert rmg.axis_units == ('mm', 'cm')


def test_name_setter_too_few_units():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.axis_units = ('m', )


def test_name_setter_too_many_units():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.axis_units = ('m', 'cm', 'mm')



#!/usr/env/python

"""
flow_accum_bw.py: Implementation of the Braun & Willet (2012) stack alorithm.

Implementation of Braun & Willett (2012) algorithm for calculating drainage
area and (optionally) water discharge. Assumes each node has only one
downstream receiver. If water discharge is calculated, the result assumes
steady flow (that is, hydrologic equilibrium).

The main public function is::

    a, q, s = flow_accumulation(r)

which takes an array of receiver-node IDs, r (the nodes that "receive" the flow
from a each node; this array would be returned by the flow_routing component's
calc_flowdirs() method). It returns Numpy
arrays with the drainage area (a) and discharge (q) at each node, along with an
array (s) that contains the IDs of the nodes in downstream-to-upstream order.

If you simply want the ordered list by itself, use::

    s = make_ordered_node_array(r)

Created: GT Nov 2013
"""
import numpy
from six.moves import range
from .cfuncs import _add_to_stack

class _DrainageStack():


    """
    Implements Braun & Willett's add_to_stack function.

    The _DrainageStack() class implements Braun & Willett's add_to_stack
    function (as a method) and also keeps track of the counter (j) and the
    stack (s). It is used by the make_ordered_node_array() function.
    """

    def __init__(self, delta, D):

        """
        Initializes the _Drainage_Stack class.

        Initializes the index counter j to zero, creates the stack array s,
        and stores references to delta and D.
        """
        self.j = 0
        self.s = numpy.zeros(len(D), dtype=int)
        self.delta = delta
        self.D = D

    def add_to_stack(self, l):

        """
        Adds node l to the stack and increments the current index (j).

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.components.flow_accum.flow_accum_bw import(
        ... _DrainageStack)
        >>> delta = np.array([ 0,  0,  2,  2,  2,  6,  7,  9, 10, 10, 10])
        >>> D = np.array([0, 2, 1, 4, 5, 7, 6, 3, 8, 9])
        >>> ds = _DrainageStack(delta, D)
        >>> ds.add_to_stack(4)
        >>> ds.s
        array([4, 1, 0, 2, 5, 6, 3, 8, 7, 9])
        """
        # we invoke cython here to attempt to suppress Python's RecursionLimit
        self.j = _add_to_stack(l, self.j, self.s, self.delta, self.D)


def _make_number_of_donors_array(r):

    """Number of donors for each node.

    Creates and returns an array containing the number of donors for each node.

    Parameters
    ----------
    r : ndarray
        ID of receiver for each node.

    Returns
    -------
    ndarray
        Number of donors for each node.

    Examples
    --------
    The example below is from Braun and Willett (2012); nd corresponds to their
    d_i in Table 1.

    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_bw import(
    ... _make_number_of_donors_array)
    >>> r = np.array([2, 5, 2, 7, 5, 5, 6, 5, 7, 8]) - 1
    >>> nd = _make_number_of_donors_array(r)
    >>> nd
    array([0, 2, 0, 0, 4, 1, 2, 1, 0, 0])
    """
    # Vectorized, DEJH, 5/20/14
#    np = len(r)
#    nd = numpy.zeros(np, dtype=int)
#    for i in range(np):
#        nd[r[i]] += 1

    nd = numpy.zeros(r.size, dtype=int)
    max_index = numpy.max(r)
    nd[:(max_index + 1)] = numpy.bincount(r)
    return nd


def _make_delta_array(nd):

    r"""
    Delta array.

    Creates and returns the "delta" array, which is a list containing, for each
    node, the array index where that node's donor list begins.

    Parameters
    ----------
    nd : ndarray of int
        Number of donors for each node

    Returns
    -------
    ndarray of int
        Delta array

    Examples
    --------
    The example below is from Braun and Willett (2012), and represents
    \delta_i in their Table 1. Here, the numbers are all one less than in their
    table because here we number indices from 0 rather than 1.

    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_bw import(
    ... _make_delta_array)
    >>> nd = np.array([0, 2, 0, 0, 4, 1, 2, 1, 0, 0])
    >>> delta = _make_delta_array(nd)
    >>> delta
    array([ 0,  0,  2,  2,  2,  6,  7,  9, 10, 10, 10])
    """
    #np = len(nd)
    #delta = numpy.zeros(np+1, dtype=int)
    #delta[np] = np   # not np+1 as in B&W because here we number from 0
    #for i in range(np-1, -1, -1):
    #    delta[i] = delta[i+1] - nd[i]
    #return delta

    #DEJH efficient delooping (only a small gain)
    np = len(nd)
    delta = numpy.zeros(np+1, dtype=int)
    delta.fill(np)
    delta[-2::-1] -= numpy.cumsum(nd[::-1])
    return delta

def _make_array_of_donors(r, delta):

    """
    Creates and returns an array containing the IDs of donors for each node.

    Essentially, the array is a series of lists (not in the Python list object
    sense) of IDs for each node. See Braun & Willett (2012) for details.

    The example below is from Braun & Willett (2012), and produces D_i in their
    Table 1 (except that here the ID numbers are one less, because we number
    indices from zero).

    Vectorized - inefficiently! - DEJH, 5/20/14

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_bw import(
    ... _make_array_of_donors)
    >>> r = np.array([2, 5, 2, 7, 5, 5, 6, 5, 7, 8])-1
    >>> delta = np.array([ 0,  0,  2,  2,  2,  6,  7,  9, 10, 10, 10])
    >>> D = _make_array_of_donors(r, delta)
    >>> D
    array([0, 2, 1, 4, 5, 7, 6, 3, 8, 9])
    """
    np = len(r)
    w = numpy.zeros(np, dtype=int)
    D = numpy.zeros(np, dtype=int)

    for i in range(np):
        ri = r[i]
        D[delta[ri]+w[ri]] = i
        w[ri] += 1

    return D

    #DEJH notes that for reasons he's not clear on, this looped version is
    #actually much slower!
    #D = numpy.zeros(np, dtype=int)
    #wri_fin = numpy.bincount(r)
    #wri_fin_nz = wri_fin.nonzero()[0]
    #wri_fin_nz_T = wri_fin_nz.reshape((wri_fin_nz.size,1))
    #logical = numpy.tile(r,(wri_fin_nz.size,1))==wri_fin_nz_T
    #cum_logical = numpy.cumsum(logical, axis=1)
    #wri = numpy.sum(numpy.where(logical, cum_logical-1,0) ,axis=0)
    #D_index = delta[r] + wri
    #D[D_index] = numpy.arange(r.size)
    #return D


def make_ordered_node_array(receiver_nodes):

    """Create an array of node IDs that is arranged in order from.

    Creates and returns an array of node IDs that is arranged in order from
    downstream to upstream.

    The lack of a leading underscore is meant to signal that this operation
    could be useful outside of this module!

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum import make_ordered_node_array
    >>> r = np.array([2, 5, 2, 7, 5, 5, 6, 5, 7, 8])-1
    >>> s = make_ordered_node_array(r)
    >>> s
    array([4, 1, 0, 2, 5, 6, 3, 8, 7, 9])
    """
    node_id = numpy.arange(receiver_nodes.size)
    baselevel_nodes = numpy.where(node_id==receiver_nodes)[0]
    nd = _make_number_of_donors_array(receiver_nodes)
    delta = _make_delta_array(nd)
    D = _make_array_of_donors(receiver_nodes, delta)
    dstack = _DrainageStack(delta, D)
    add_it = dstack.add_to_stack
    for k in baselevel_nodes:
        add_it(k) #don't think this is a bottleneck, so no C++

    return dstack.s


def find_drainage_area_and_discharge(s, r, node_cell_area=1.0, runoff=1.0,
                                     boundary_nodes=None):

    """Calculate the drainage area and water discharge at each node.

    Parameters
    ----------
    s : ndarray of int
        Ordered (downstream to upstream) array of node IDs
    r : ndarray of int
        Receiver IDs for each node
    node_cell_area : float or ndarray
        Cell surface areas for each node. If it's an array, must have same
        length as s (that is, the number of nodes).
    runoff : float or ndarray
        Local runoff rate at each cell (in water depth per time). If it's an
        array, must have same length as s (that is, the number of nodes).

    Returns
    -------
    tuple of ndarray
        drainage area and discharge

    Notes
    -----
    -  If node_cell_area not given, the output drainage area is equivalent
       to the number of nodes/cells draining through each point, including
       the local node itself.
    -  Give node_cell_area as a scalar when using a regular raster grid.
    -  If runoff is not given, the discharge returned will be the same as
       drainage area (i.e., drainage area times unit runoff rate).
    -  If using an unstructured Landlab grid, make sure that the input
       argument for node_cell_area is the cell area at each NODE rather than
       just at each CELL. This means you need to include entries for the
       perimeter nodes too. They can be zeros.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum import (
    ...     find_drainage_area_and_discharge)
    >>> r = np.array([2, 5, 2, 7, 5, 5, 6, 5, 7, 8])-1
    >>> s = np.array([4, 1, 0, 2, 5, 6, 3, 8, 7, 9])
    >>> a, q = find_drainage_area_and_discharge(s, r)
    >>> a
    array([  1.,   3.,   1.,   1.,  10.,   4.,   3.,   2.,   1.,   1.])
    >>> q
    array([  1.,   3.,   1.,   1.,  10.,   4.,   3.,   2.,   1.,   1.])
    """
    # Number of points
    np = len(s)

    # Initialize the drainage_area and discharge arrays. Drainage area starts
    # out as the area of the cell in question, then (unless the cell has no
    # donors) grows from there. Discharge starts out as the cell's local runoff
    # rate times the cell's surface area.
    drainage_area = numpy.zeros(np, dtype=int) + node_cell_area
    discharge = numpy.zeros(np, dtype=int) + node_cell_area*runoff

    # Optionally zero out drainage area and discharge at boundary nodes
    if boundary_nodes is not None:
        drainage_area[boundary_nodes] = 0
        discharge[boundary_nodes] = 0

    # Iterate backward through the list, which means we work from upstream to
    # downstream.
    for i in range(np-1, -1, -1):
        donor = s[i]
        recvr = r[donor]
        if donor != recvr:
            drainage_area[recvr] += drainage_area[donor]
            discharge[recvr] += discharge[donor]

    return drainage_area, discharge


def flow_accumulation(receiver_nodes, node_cell_area=1.0,
                      runoff_rate=1.0, boundary_nodes=None):

    """Calculate drainage area and (steady) discharge.

    Calculates and returns the drainage area and (steady) discharge at each
    node, along with a downstream-to-upstream ordered list (array) of node IDs.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum import flow_accumulation
    >>> r = np.array([2, 5, 2, 7, 5, 5, 6, 5, 7, 8])-1
    >>> a, q, s = flow_accumulation(r)
    >>> a
    array([  1.,   3.,   1.,   1.,  10.,   4.,   3.,   2.,   1.,   1.])
    >>> q
    array([  1.,   3.,   1.,   1.,  10.,   4.,   3.,   2.,   1.,   1.])
    >>> s
    array([4, 1, 0, 2, 5, 6, 3, 8, 7, 9])
    """

    s = make_ordered_node_array(receiver_nodes)
    #Note that this ordering of s DOES INCLUDE closed nodes. It really shouldn't!
    #But as we don't have a copy of the grid accessible here, we'll solve this
    #problem as part of route_flow_dn.

    a, q = find_drainage_area_and_discharge(s, receiver_nodes, node_cell_area,
                                            runoff_rate, boundary_nodes)

    return a, q, s


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#! /usr/env/python

"""
flow_director.py provides a private class to help create FlowDirectors.

Provides the _FlowDirector component which does grid type testing, adds the
surface over which flow will be routed to the component, and sets up part of
the boundary condition testing.
"""

from __future__ import print_function
from landlab import Component
from landlab import RasterModelGrid  # for type tests
from landlab.utils.return_array import return_array_at_node
import six


class _FlowDirector(Component):

    """
    Private class for creating components to calculate flow directions.

    This class is not meant to be used directly in modeling efforts.
    Instead it has the functionality that all flow direction calculators need
    to initialize and check boundary conditions.

    The primary method of this class, :func:`run_one_step` is not implemented.

    Parameters
    ----------
    grid : ModelGrid
        A grid.
    surface : field name at node or array of length node
        The surface to direct flow across.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.flow_director.flow_director import(
    ... _FlowDirector)
    >>> mg = RasterModelGrid((3,3), spacing=(1, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> _ = mg.add_field('topographic__elevation',
    ...                  mg.node_x + mg.node_y,
    ...                  at = 'node')
    >>> fd = _FlowDirector(mg, 'topographic__elevation')
    >>> fd.surface_values
    array([ 0.,  1.,  2.,  1.,  2.,  3.,  2.,  3.,  4.])
    >>> list(mg.at_node.keys())
    ['topographic__elevation']

    _FlowDirector also works if you pass it an array instead of a field name.
    >>> import numpy as np
    >>> mg = RasterModelGrid((3,3), spacing=(1, 1))
    >>> z = np.array([ 0.,  1.,  2.,  1.,  2.,  3.,  2.,  3.,  4.])
    >>> fd = _FlowDirector(mg, z)
    >>> fd.surface_values
    array([ 0.,  1.,  2.,  1.,  2.,  3.,  2.,  3.,  4.])
    """

    _name = '_FlowDirector'

    def __init__(self, grid, surface):
        """Initialize the _FlowDirector class."""
        # We keep a local reference to the grid
        super(_FlowDirector, self).__init__(grid)

        self._grid = grid
        self._bc_set_code = self.grid.bc_set_code

        # set up the grid type testing
        self._is_raster = isinstance(self._grid, RasterModelGrid)
        if not self._is_raster:
            self.method = None

        # save elevations as class properites.
        self.surface = surface
        self.surface_values = return_array_at_node(grid, surface)

    def _changed_surface(self):
        """Check if the surface values have changed.

        If the surface values are stored as a field, it is important to check
        if they have changed since the component was instantiated.
        """
        if isinstance(self.surface, six.string_types):
            self.surface_values = return_array_at_node(self._grid, self.surface)

    def run_one_step(self):
        """run_one_step is not implemented for this component."""
        raise NotImplementedError('run_one_step()')

if __name__ == '__main__':
    import doctest
    doctest.testmod()



import pytest

from landlab import RasterModelGrid
from landlab.components import LandslideProbability


@pytest.fixture
def ls_prob():
    grid = RasterModelGrid((20, 20), spacing=10e0)
    grid.add_zeros('topographic__slope', at='node', dtype=float)
    return LandslideProbability(grid)



import numpy as np
cimport numpy as np
cimport cython


DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def fill_links_at_patch(np.ndarray[DTYPE_t, ndim=1] links_at_patch,
                        np.ndarray[DTYPE_t, ndim=1] offset_to_patch,
                        np.ndarray[DTYPE_t, ndim=2] out):
    cdef int i
    cdef int link
    cdef int patch
    cdef int offset
    cdef int n_links
    cdef int n_patches = len(offset_to_patch) - 1

    for patch in range(n_patches):
        offset = offset_to_patch[patch]
        n_links = offset_to_patch[patch + 1] - offset

        link = 0
        for i in range(offset, offset + n_links):
          out[patch, link] = links_at_patch[i]
          link += 1
